from flask import abort
from flask_jwt_extended import jwt_required

from sqlalchemy import func

from zou.app.utils import events

from zou.app.blueprints.crud.base import BaseModelsResource, BaseModelResource
from zou.app.services import user_service, tasks_service

from zou.app.models.time_spent import TimeSpent
from zou.app.models.task import Task


class TimeSpentsResource(BaseModelsResource):
    def __init__(self):
        BaseModelsResource.__init__(self, TimeSpent)

    @jwt_required()
    def get(self):
        """
        Get time spents
        ---
        tags:
          - Crud
        description: Retrieve all time spent records. Supports filtering
          via query parameters and pagination. Supports date range
          filtering with start_date and end_date.
        parameters:
          - in: query
            name: page
            required: false
            schema:
              type: integer
            example: 1
            description: Page number for pagination
          - in: query
            name: limit
            required: false
            schema:
              type: integer
            example: 50
            description: Number of results per page
          - in: query
            name: relations
            required: false
            schema:
              type: boolean
            default: false
            example: false
            description: Whether to include relations
          - in: query
            name: start_date
            required: false
            schema:
              type: string
              format: date
            example: "2024-01-01"
            description: Start date for date range filter
          - in: query
            name: end_date
            required: false
            schema:
              type: string
              format: date
            example: "2024-01-31"
            description: End date for date range filter
        responses:
            200:
              description: Time spent records retrieved successfully
              content:
                application/json:
                  schema:
                    oneOf:
                      - type: array
                        items:
                          type: object
                      - type: object
                        properties:
                          data:
                            type: array
                            items:
                              type: object
                            example: []
                          total:
                            type: integer
                            example: 100
                          nb_pages:
                            type: integer
                            example: 2
                          limit:
                            type: integer
                            example: 50
                          offset:
                            type: integer
                            example: 0
                          page:
                            type: integer
                            example: 1
            400:
              description: Invalid filter format or query error
        """
        return super().get()

    @jwt_required()
    def post(self):
        """
        Create time spent
        ---
        tags:
          - Crud
        description: Create a new time spent record with data provided
          in the request body. JSON format is expected. Updates task
          duration automatically.
        requestBody:
          required: true
          content:
            application/json:
              schema:
                type: object
                required:
                  - task_id
                  - person_id
                  - date
                  - duration
                properties:
                  task_id:
                    type: string
                    format: uuid
                    example: a24a6ea4-ce75-4665-a070-57453082c25
                  person_id:
                    type: string
                    format: uuid
                    example: b24a6ea4-ce75-4665-a070-57453082c25
                  date:
                    type: string
                    format: date
                    example: "2024-01-15"
                  duration:
                    type: number
                    example: 8.5
        responses:
            201:
              description: Time spent record created successfully
              content:
                application/json:
                  schema:
                    type: object
                    properties:
                      id:
                        type: string
                        format: uuid
                        example: a24a6ea4-ce75-4665-a070-57453082c25
                      task_id:
                        type: string
                        format: uuid
                        example: b24a6ea4-ce75-4665-a070-57453082c25
                      person_id:
                        type: string
                        format: uuid
                        example: c24a6ea4-ce75-4665-a070-57453082c25
                      date:
                        type: string
                        format: date
                        example: "2024-01-15"
                      duration:
                        type: number
                        example: 8.5
                      created_at:
                        type: string
                        format: date-time
                        example: "2024-01-15T10:30:00Z"
                      updated_at:
                        type: string
                        format: date-time
                        example: "2024-01-15T10:30:00Z"
            400:
              description: Invalid data format or validation error
        """
        return super().post()

    def apply_filters(self, query, options):
        query = super(TimeSpentsResource, self).apply_filters(query, options)
        start_date = options.get("start_date", None)
        end_date = options.get("end_date", None)
        if not start_date and not end_date:
            return query

        if None in [start_date, end_date]:
            abort(
                400,
                "If querying for a range of dates, both a `start_date` and an "
                "`end_date` must be given.",
            )

        return query.filter(
            self.model.date.between(
                func.cast(start_date, TimeSpent.date.type)
            ),
            func.cast(end_date, TimeSpent.date.type),
        )

    def post_creation(self, instance):
        task_id = str(instance.task_id)
        task = Task.get(task_id)
        task.duration = sum(
            time_spent.duration
            for time_spent in TimeSpent.get_all_by(task_id=task_id)
        )
        task.save()
        tasks_service.clear_task_cache(task_id)
        events.emit(
            "task:update",
            {"task_id": task_id},
            project_id=str(task.project_id),
        )
        return instance.serialize()


class TimeSpentResource(BaseModelResource):
    def __init__(self):
        BaseModelResource.__init__(self, TimeSpent)

    def check_delete_permissions(self, instance_dict):
        return user_service.check_time_spent_access(
            instance_dict["task_id"], instance_dict["person_id"]
        )

    @jwt_required()
    def get(self, instance_id):
        """
        Get time spent
        ---
        tags:
          - Crud
        description: Retrieve a time spent record by its ID and return
          it as a JSON object. Supports including relations.
        parameters:
          - in: path
            name: instance_id
            required: true
            schema:
              type: string
              format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
          - in: query
            name: relations
            required: false
            schema:
              type: boolean
            default: true
            example: true
            description: Whether to include relations
        responses:
            200:
              description: Time spent record retrieved successfully
              content:
                application/json:
                  schema:
                    type: object
                    properties:
                      id:
                        type: string
                        format: uuid
                        example: a24a6ea4-ce75-4665-a070-57453082c25
                      task_id:
                        type: string
                        format: uuid
                        example: b24a6ea4-ce75-4665-a070-57453082c25
                      person_id:
                        type: string
                        format: uuid
                        example: c24a6ea4-ce75-4665-a070-57453082c25
                      date:
                        type: string
                        format: date
                        example: "2024-01-15"
                      duration:
                        type: number
                        example: 8.5
                      created_at:
                        type: string
                        format: date-time
                        example: "2024-01-15T10:30:00Z"
                      updated_at:
                        type: string
                        format: date-time
                        example: "2024-01-15T10:30:00Z"
            400:
              description: Invalid ID format or query error
        """
        return super().get(instance_id)

    @jwt_required()
    def put(self, instance_id):
        """
        Update time spent
        ---
        tags:
          - Crud
        description: Update a time spent record with data provided in
          the request body. JSON format is expected. Updates task
          duration automatically.
        parameters:
          - in: path
            name: instance_id
            required: true
            schema:
              type: string
              format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
        requestBody:
          required: true
          content:
            application/json:
              schema:
                type: object
                properties:
                  date:
                    type: string
                    format: date
                    example: "2024-01-16"
                  duration:
                    type: number
                    example: 7.5
        responses:
            200:
              description: Time spent record updated successfully
              content:
                application/json:
                  schema:
                    type: object
                    properties:
                      id:
                        type: string
                        format: uuid
                        example: a24a6ea4-ce75-4665-a070-57453082c25
                      task_id:
                        type: string
                        format: uuid
                        example: b24a6ea4-ce75-4665-a070-57453082c25
                      person_id:
                        type: string
                        format: uuid
                        example: c24a6ea4-ce75-4665-a070-57453082c25
                      date:
                        type: string
                        format: date
                        example: "2024-01-16"
                      duration:
                        type: number
                        example: 7.5
                      created_at:
                        type: string
                        format: date-time
                        example: "2024-01-15T10:30:00Z"
                      updated_at:
                        type: string
                        format: date-time
                        example: "2024-01-15T11:00:00Z"
            400:
              description: Invalid data format or validation error
        """
        return super().put(instance_id)

    @jwt_required()
    def delete(self, instance_id):
        """
        Delete time spent
        ---
        tags:
          - Crud
        description: Delete a time spent record by its ID. Returns
          empty response on success. Updates task duration
          automatically.
        parameters:
          - in: path
            name: instance_id
            required: true
            schema:
              type: string
              format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            204:
              description: Time spent record deleted successfully
            400:
              description: Integrity error or cannot delete
        """
        return super().delete(instance_id)

    def post_update(self, instance_dict, data):
        task = Task.get(instance_dict["task_id"])
        task.duration = sum(
            time_spent.duration
            for time_spent in TimeSpent.get_all_by(
                task_id=instance_dict["task_id"]
            )
        )
        task.save()
        tasks_service.clear_task_cache(instance_dict["task_id"])
        events.emit(
            "task:update",
            {"task_id": instance_dict["task_id"]},
            project_id=str(task.project_id),
        )
        return instance_dict

    def post_delete(self, instance_dict):
        task = Task.get(instance_dict["task_id"])
        task.duration = sum(
            time_spent.duration
            for time_spent in TimeSpent.get_all_by(
                task_id=instance_dict["task_id"]
            )
        )
        task.save()
        tasks_service.clear_task_cache(instance_dict["task_id"])
        events.emit(
            "task:update",
            {"task_id": instance_dict["task_id"]},
            project_id=str(task.project_id),
        )
        return instance_dict



import base64
import logging

import orjson as json
import os
import zlib

from flask import current_app
from operator import itemgetter
from pathlib import Path
from shutil import copyfile
from zipfile import ZipFile

from flask_fs.errors import FileNotFound
from slugify import slugify
from sqlalchemy import or_
from sqlalchemy.orm import joinedload

from zou.app import config
from zou.app.stores import config_store, file_store

from zou.app.models.build_job import BuildJob
from zou.app.models.playlist import Playlist
from zou.app.models.preview_file import PreviewFile
from zou.app.models.task import Task
from zou.app.models.task_type import TaskType

from zou.utils import movie
from zou.app.utils import fields, events, remote_job, emails
from zou.app.utils import query as query_utils
from zou.app.stores.redis_lock import with_playlist_lock

from zou.app.services import (
    assets_service,
    base_service,
    entities_service,
    files_service,
    preview_files_service,
    projects_service,
    shots_service,
    tasks_service,
    names_service,
    persons_service,
    templates_service,
)

from zou.app.services.exception import (
    BuildJobNotFoundException,
    PlaylistNotFoundException,
)

logger = logging.getLogger(__name__)

# Page size for playlist list endpoints
PLAYLISTS_PAGE_SIZE = 20

# Scalar attributes for playlist list dict (avoids serializing heavy shots JSON)
_PLAYLIST_LIST_ATTRS = (
    "id",
    "name",
    "project_id",
    "episode_id",
    "task_type_id",
    "for_client",
    "for_entity",
    "is_for_all",
    "created_at",
    "updated_at",
    "created_by",
)


def _apply_playlist_pagination(query, page, sort_by, model=Playlist):
    """Apply sort, normalize page, and apply limit/offset to the query."""
    query = query_utils.apply_sort_by(model, query, sort_by)
    if page < 1:
        page = 1
    offset = (page - 1) * PLAYLISTS_PAGE_SIZE
    return query.limit(PLAYLISTS_PAGE_SIZE).offset(offset)


def all_playlists_for_project(
    project_id,
    for_client=False,
    page=1,
    sort_by="updated_at",
    task_type_id=None,
):
    """
    Return all playlists created for given project.
    """
    result = []
    query = Playlist.query.filter(Playlist.project_id == project_id)
    if for_client:
        query = query.filter(Playlist.for_client)

    if task_type_id is not None and len(task_type_id) > 0:
        query = query.filter(Playlist.task_type_id == task_type_id)

    query = _apply_playlist_pagination(query, page, sort_by)
    playlists = query.all()
    for playlist in playlists:
        playlist_dict = build_playlist_dict(playlist)
        result.append(playlist_dict)
    return result


def all_playlists_for_episode(
    project_id,
    episode_id,
    for_client=False,
    page=1,
    sort_by="updated_at",
    task_type_id=None,
):
    """
    Return all playlists created for given episode.
    """
    result = []
    query = Playlist.query
    if for_client:
        query = query.filter(Playlist.for_client)

    if task_type_id is not None and len(task_type_id) > 0:
        query = query.filter(Playlist.task_type_id == task_type_id)

    if episode_id == "main":
        query = (
            query.filter(Playlist.episode_id == None)
            .filter(Playlist.project_id == project_id)
            .filter(
                or_(Playlist.is_for_all is None, Playlist.is_for_all == False)
            )
        )
    elif episode_id == "all":
        query = (
            query.filter(Playlist.episode_id == None)
            .filter(Playlist.project_id == project_id)
            .filter(Playlist.is_for_all == True)
        )
    else:
        query = query.filter(Playlist.episode_id == episode_id)

    query = _apply_playlist_pagination(query, page, sort_by)
    playlists = query.all()
    for playlist in playlists:
        playlist_dict = build_playlist_dict(playlist)
        result.append(playlist_dict)
    return result


def build_playlist_dict(playlist):
    """
    Build a dictionary of a simplified version of the playlist. It just takes
    the information needed for displaying the list of playlists.
    Does not mutate the playlist model. Uses scalar attributes only to avoid
    serializing the potentially large shots JSON column.
    """
    out = {"type": "Playlist"}
    for attr in _PLAYLIST_LIST_ATTRS:
        if hasattr(playlist, attr):
            out[attr] = fields.serialize_value(getattr(playlist, attr))
    if playlist.for_entity is None:
        out["for_entity"] = "shot"
    first_shot_preview_file_id = get_first_shot_preview_file_id(playlist)
    if first_shot_preview_file_id is not None:
        out["first_preview_file_id"] = first_shot_preview_file_id
    return out


def get_first_shot_preview_file_id(playlist):
    """Return the first shot's preview_file_id if any, else None."""
    if not playlist.shots or not isinstance(playlist.shots, list):
        return None
    first = playlist.shots[0]
    return first.get("preview_file_id") if isinstance(first, dict) else None


def get_playlist_with_preview_file_revisions(playlist_id):
    """
    Return given playlist. Shot list is augmented with all previews available
    for a given shot.
    """
    # Eager load build_jobs to avoid N+1 when building build_jobs list
    playlist = (
        Playlist.query.options(joinedload(Playlist.build_jobs))
        .filter(Playlist.id == playlist_id)
        .first()
    )
    if playlist is None:
        raise PlaylistNotFoundException()
    playlist_dict = playlist.serialize()
    playlist_dict = _add_build_job_infos_to_playlist_dict(
        playlist, playlist_dict
    )

    if playlist_dict["shots"] is None:
        playlist_dict["shots"] = []
    playlist_dict, preview_file_map = set_preview_files_for_entities(
        playlist_dict
    )

    for shot in playlist_dict["shots"]:
        try:
            preview_file = preview_file_map.get(shot["preview_file_id"], None)
            if preview_file is not None:
                shot["preview_file_id"] = preview_file["id"]
                shot["preview_file_extension"] = preview_file["extension"]
                shot["preview_file_revision"] = preview_file["revision"]
                shot["preview_file_width"] = preview_file["width"]
                shot["preview_file_height"] = preview_file["height"]
                shot["preview_file_duration"] = preview_file["duration"]
                shot["preview_file_status"] = preview_file["status"]
                shot["preview_file_annotations"] = preview_file["annotations"]
                shot["preview_file_task_id"] = preview_file["task_id"]
                shot["preview_file_previews"] = preview_file.get(
                    "previews", []
                )
            else:
                del shot["preview_file_id"]
        except Exception as e:
            logger.warning(
                "Failed to enrich shot with preview file: shot_id=%s, error=%s",
                shot.get("id"),
                e,
            )
    return playlist_dict


def _add_build_job_infos_to_playlist_dict(playlist, playlist_dict):
    playlist_dict["build_jobs"] = []
    for build_job in reversed(playlist.build_jobs):
        playlist_dict["build_jobs"].append(build_job.present())
    return playlist_dict


def set_preview_files_for_entities(playlist_dict):
    """
    Retrieve all preview files related to entities listed in given playlist.
    Add to each entity a dict with task as keys and preview list as values.
    """
    entity_ids = []
    for entity in playlist_dict["shots"]:
        if "id" not in entity:
            entity_id = entity.get("shot_id", entity.get("entity_id", None))
            if entity_id is not None:
                entity_ids.append(entity_id)
                entity["id"] = entity_id
        else:
            entity_ids.append(entity["id"])
    previews = {}
    preview_file_map = {}

    preview_files = (
        PreviewFile.query.join(Task)
        .join(TaskType)
        .filter(Task.entity_id.in_(entity_ids))
        .order_by(TaskType.priority.desc())
        .order_by(TaskType.name)
        .order_by(PreviewFile.revision.desc())
        .order_by(PreviewFile.position)
        .order_by(PreviewFile.created_at)
        .add_column(Task.task_type_id)
        .add_column(Task.entity_id)
        .all()
    )

    for preview_file, task_type_id, entity_id in preview_files:
        entity_id = str(entity_id)
        task_type_id = str(task_type_id)
        if entity_id not in previews:
            previews[entity_id] = {}

        if task_type_id not in previews[entity_id]:
            previews[entity_id][task_type_id] = []

        task_id = str(preview_file.task_id)
        preview_file_id = str(preview_file.id)

        light_preview_file = {
            "id": preview_file_id,
            "revision": preview_file.revision,
            "extension": preview_file.extension,
            "width": preview_file.width,
            "height": preview_file.height,
            "duration": float(preview_file.duration or 0),
            "status": str(preview_file.status),
            "annotations": preview_file.annotations,
            "created_at": fields.serialize_value(preview_file.created_at),
            "task_id": task_id,
        }  # Do not add too much field to avoid building too big responses
        previews[entity_id][task_type_id].append(light_preview_file)
        preview_file_map[preview_file_id] = light_preview_file

    for entity_id in previews.keys():
        for task_type_id in previews[entity_id].keys():
            previews[entity_id][task_type_id] = mix_preview_file_revisions(
                previews[entity_id][task_type_id]
            )

    for entity in playlist_dict["shots"]:
        if str(entity["id"]) in previews:
            entity["preview_files"] = previews[str(entity["id"])]
        else:
            entity["preview_files"] = []

    return (fields.serialize_value(playlist_dict), preview_file_map)


def get_preview_files_for_entity(entity_id):
    """
    Get all preview files available for given shot.
    """
    previews = {}
    query = (
        Task.query.filter_by(entity_id=entity_id)
        .add_columns(
            PreviewFile.id,
            PreviewFile.revision,
            PreviewFile.position,
            PreviewFile.original_name,
            PreviewFile.extension,
            PreviewFile.width,
            PreviewFile.height,
            PreviewFile.duration,
            PreviewFile.status,
            PreviewFile.annotations,
            PreviewFile.created_at,
            PreviewFile.task_id,
        )
        .join(PreviewFile, Task.id == PreviewFile.task_id)
        .join(TaskType)
        .order_by(TaskType.priority.desc())
        .order_by(TaskType.name)
        .order_by(PreviewFile.revision.desc())
        .order_by(PreviewFile.created_at)
    )

    task_previews = {}
    for (
        task,
        preview_file_id,
        preview_file_revision,
        preview_file_position,
        preview_file_original_name,
        preview_file_extension,
        preview_file_width,
        preview_file_height,
        preview_file_duration,
        preview_file_status,
        preview_file_annotations,
        preview_file_created_at,
        preview_file_task_id,
    ) in query.all():
        task_id = str(task.id)
        if task_id not in task_previews:
            task_previews[task_id] = []
        task_previews[task_id].append(
            fields.serialize_dict(
                {
                    "id": preview_file_id,
                    "revision": preview_file_revision,
                    "position": preview_file_position,
                    "original_name": preview_file_original_name,
                    "extension": preview_file_extension,
                    "width": preview_file_width,
                    "height": preview_file_height,
                    "duration": float(preview_file_duration or 0),
                    "status": preview_file_status,
                    "annotations": preview_file_annotations,
                    "created_at": preview_file_created_at,
                    "task_id": task_id,
                    "task_type_id": str(task.task_type_id),
                }
            )
        )

    for task_id in task_previews.keys():
        preview_files = task_previews[task_id]
        task_type_id = task_previews[task_id][0]["task_type_id"]

        if len(preview_files) > 0:
            preview_files = mix_preview_file_revisions(preview_files)
            previews[task_type_id] = [
                {
                    "id": preview_file["id"],
                    "revision": preview_file["revision"],
                    "original_name": preview_file["original_name"],
                    "extension": preview_file["extension"],
                    "width": preview_file["width"],
                    "height": preview_file["height"],
                    "duration": preview_file["duration"],
                    "status": preview_file["status"],
                    "annotations": preview_file["annotations"],
                    "previews": preview_file["previews"],
                    "created_at": preview_file["created_at"],
                    "task_id": preview_file["task_id"],
                }
                for preview_file in preview_files
            ]  # Do not add too much field to avoid building too big responses
    return previews


def mix_preview_file_revisions(preview_files):
    """
    The goal here is to group preview files with same revision in a single
    preview file, which encapsulates other preview_files.
    """
    revision_map = {}
    result = []
    for preview_file in preview_files:
        revision = preview_file["revision"]
        if revision not in revision_map:
            preview_file["previews"] = []
            revision_map[revision] = preview_file
            result.append(preview_file)
        else:
            parent_preview_file = revision_map[revision]
            parent_preview_file["previews"].append(preview_file)
    return result


def get_playlist_raw(playlist_id):
    """
    Return given playlist as active record.
    """
    return base_service.get_instance(
        Playlist, playlist_id, PlaylistNotFoundException
    )


def get_playlist(playlist_id):
    """
    Return given playlist as a dict.
    """
    return get_playlist_raw(playlist_id).serialize()


def add_entity_to_playlist(playlist_id, entity_id, preview_file_id=None):
    """
    Atomically append an entity to the playlist shots list.
    This function is designed to be safe under concurrent access.

    Uses a Redis-based distributed lock to prevent race conditions
    when multiple processes try to add entities to the same playlist
    concurrently. Falls back to database-level locking if Redis is
    unavailable.
    """
    entity_id_str = str(entity_id)
    with with_playlist_lock(
        playlist_id, timeout=30, wait_timeout=35
    ) as _acquired:
        playlist_dict = _add_entity_to_playlist_db(
            playlist_id, entity_id_str, preview_file_id
        )

    events.emit(
        "playlist:add_entity",
        {
            "playlist_id": playlist_dict["id"],
            "entity_id": entity_id_str,
            "preview_file_id": (
                str(preview_file_id) if preview_file_id is not None else None
            ),
        },
        project_id=playlist_dict["project_id"],
    )

    return playlist_dict


def _add_entity_to_playlist_db(playlist_id, entity_id_str, preview_file_id):
    """
    Internal helper function to add an entity to playlist in the database.
    Assumes the caller has acquired appropriate locking (Redis).
    """
    playlist = Playlist.get(playlist_id)
    shots = list(playlist.shots or [])

    if not any(shot.get("entity_id") == entity_id_str for shot in shots):
        shot = {
            "entity_id": entity_id_str,
        }
        if preview_file_id is not None:
            shot["preview_file_id"] = str(preview_file_id)
        shots.append(shot)
        playlist.shots = shots
        playlist.update({"shots": shots})

    return playlist.serialize()


def playlist_previews(shots, only_movies=False):
    """
    Retrieve all preview id and extension for the given shots.
    Uses a single batch query instead of N queries (avoids N+1).
    """
    preview_file_ids = []
    for entity in shots:
        if not isinstance(entity, dict):
            continue
        pf_id = entity.get("preview_file_id")
        if pf_id is not None and len(str(pf_id).strip()) > 0:
            preview_file_ids.append(str(pf_id))
    if not preview_file_ids:
        return []

    # Single query for all preview files (batch fetch)
    preview_models = PreviewFile.query.filter(
        PreviewFile.id.in_(preview_file_ids)
    ).all()
    id_to_preview = {
        str(pf.id): {"id": str(pf.id), "extension": pf.extension}
        for pf in preview_models
    }

    # Preserve order of shots and apply only_movies filter
    result = []
    for entity in shots:
        if not isinstance(entity, dict):
            continue
        pf_id = entity.get("preview_file_id")
        if not pf_id:
            continue
        preview = id_to_preview.get(str(pf_id))
        if preview is None:
            continue
        if only_movies and preview["extension"] != "mp4":
            continue
        result.append(preview)
    return result


def retrieve_playlist_tmp_files(preview_files, full=False):
    """
    Retrieve all files for a given playlist into the temporary folder.
    """
    file_paths = []
    for preview_file in preview_files:
        if full:
            preview_file = files_service.get_preview_file(preview_file["id"])
            sub_preview_files = (
                preview_files_service.get_preview_files_for_revision(
                    preview_file["task_id"], preview_file["revision"]
                )
            )
            for preview_file in sub_preview_files:
                tmp_file_path, file_name = retrieve_playlist_tmp_file(
                    preview_file
                )
                file_paths.append((tmp_file_path, file_name))
        else:
            tmp_file_path, file_name = retrieve_playlist_tmp_file(preview_file)
            file_paths.append((tmp_file_path, file_name))
    return file_paths


def retrieve_playlist_tmp_file(preview_file):
    if preview_file["extension"] == "mp4":
        get_path_func = file_store.get_local_movie_path
        open_func = file_store.open_movie
        exists_func = file_store.exists_movie
        prefix = "previews"
    elif preview_file["extension"] == "png":
        get_path_func = file_store.get_local_picture_path
        open_func = file_store.open_picture
        exists_func = file_store.exists_picture
        prefix = "original"
    else:
        get_path_func = file_store.get_local_file_path
        open_func = file_store.open_file
        exists_func = file_store.exists_file
        prefix = "previews"

    if config.FS_BACKEND == "local":
        file_path = get_path_func(prefix, preview_file["id"])
    else:
        file_path = os.path.join(
            config.TMP_DIR,
            "cache-previews-%s.%s"
            % (preview_file["id"], preview_file["extension"]),
        )
        if not os.path.exists(file_path) or os.path.getsize(file_path) == 0:
            if exists_func(prefix, preview_file["id"]):
                with open(file_path, "wb") as tmp_file:
                    try:
                        for chunk in open_func(prefix, preview_file["id"]):
                            tmp_file.write(chunk)
                    except FileNotFound:
                        pass
    file_name = names_service.get_preview_file_name(preview_file["id"])
    tmp_file_path = os.path.join(config.TMP_DIR, file_name)
    copyfile(file_path, tmp_file_path)
    return tmp_file_path, file_name


def build_playlist_zip_file(playlist):
    """
    Build a zip for all files for a given playlist into the temporary folder.
    """
    previews = playlist_previews(playlist["shots"])
    tmp_file_paths = retrieve_playlist_tmp_files(previews, full=True)

    zip_file_path = get_playlist_zip_file_path(playlist)
    if os.path.exists(zip_file_path):
        os.remove(zip_file_path)
    with ZipFile(zip_file_path, "w") as zip:
        for file_path, file_name in tmp_file_paths:
            zip.write(file_path, file_name)
    return zip_file_path


def build_playlist_movie_file(playlist, job, shots, params, full, remote):
    """
    Build a movie for all files for a given playlist into the temporary folder.
    """
    success = False
    from zou.app import app

    with app.app_context():
        try:
            previews = playlist_previews(shots, only_movies=True)
            movie_file_path = get_playlist_movie_file_path(job)
            tmp_file_paths = retrieve_playlist_tmp_files(previews)

            if tmp_file_paths:
                if not remote:
                    success = False
                    if not full:
                        success = _run_concatenation(
                            playlist,
                            job,
                            tmp_file_paths,
                            movie_file_path,
                            params,
                            movie.concat_demuxer,
                        )

                    # Try again using concat filter
                    if not success:
                        success = _run_concatenation(
                            playlist,
                            job,
                            tmp_file_paths,
                            movie_file_path,
                            params,
                            movie.concat_filter,
                        )
                else:
                    try:
                        _run_remote_job_build_playlist(
                            app, job, previews, params, movie_file_path, full
                        )
                        success = True
                    except Exception as exc:
                        app.logger.error(exc)
                        success = False

        except Exception as exc:
            app.logger.error(exc)
            success = False

        # exception will be logged by rq
        finally:
            job = end_build_job(playlist, job, success)

    if not success:
        raise Exception("Failure while building playlist %r" % playlist["id"])

    return job


def _run_concatenation(
    playlist, job, tmp_file_paths, movie_file_path, params, mode
):
    success = False
    try:
        result = movie.build_playlist_movie(
            mode, tmp_file_paths, movie_file_path, **params._asdict()
        )
        if result["success"] and os.path.exists(movie_file_path):
            file_store.add_movie("playlists", job["id"], movie_file_path)
            success = True
        if result.get("message"):
            from zou.app import app

            with app.app_context():
                app.logger.error(result["message"])
    except Exception:
        from zou.app import app

        with app.app_context():
            app.logger.error(
                "Unable to build playlist %r using %s",
                (playlist["id"], mode.__qualname__),
                exc_info=1,
            )
    return success


def _run_remote_job_build_playlist(
    app, job, previews, params, movie_file_path, full
):
    preview_ids = [
        preview["id"] for preview in previews if preview["extension"] == "mp4"
    ]
    input_bytes = zlib.compress(json.dumps(preview_ids))
    input_string = base64.b64encode(input_bytes).decode("ascii")
    params = {
        "version": "1",
        "output_filename": Path(movie_file_path).name,
        "output_key": file_store.make_key("playlists", job["id"]),
        "input": input_string,
        "width": params.width,
        "height": params.height,
        "fps": params.fps,
        "full": str(full).lower(),
    }
    nomad_job = config_store.get_nomad_playlist_job()
    remote_job.run_job(app, config, nomad_job, params)

    with open(movie_file_path, "wb") as movie_file:
        for chunk in file_store.open_movie("playlists", job["id"]):
            movie_file.write(chunk)

    return movie_file_path


def start_build_job(playlist):
    """
    clients that a new job is running.
    Register in database that a new build is running. Emits an event to notify
    """
    job = BuildJob.create(
        status="running", job_type="movie", playlist_id=playlist["id"]
    )
    events.emit(
        "build-job:new",
        {
            "build_job_id": str(job.id),
            "playlist_id": playlist["id"],
            "created_at": fields.serialize_value(job.created_at),
        },
        project_id=playlist["project_id"],
    )
    return job.serialize()


def end_build_job(playlist, job, success):
    """
    Register in database that a build is finished. Emits an event to notify
    clients that the build is done.
    """
    if success:
        status = "succeeded"
    else:
        status = "failed"

    build_job = BuildJob.get(job["id"])
    if build_job is not None:
        build_job.end(status=status)
    events.emit(
        "build-job:update",
        {
            "build_job_id": job["id"],
            "playlist_id": playlist["id"],
            "status": status,
        },
        project_id=playlist["project_id"],
    )
    if build_job is not None:
        return build_job.serialize()
    else:
        return {}


def build_playlist_job(playlist, job, shots, params, email, full, remote):
    """
    Build playlist file (concatenate all movie previews). This function is
    aimed at being run as a job in a job queue.
    """
    build_playlist_movie_file(playlist, job, shots, params, full, remote)

    # Just in case, since rq jobs which encounter an error raise an
    # exception in order to be flagged as failed.
    if job["status"] == "succeeded":
        person = persons_service.get_person_by_email_raw(email)
        organisation = persons_service.get_organisation()
        playlist_url = "%s://%s/api/data/playlists/%s/jobs/%s/build/mp4" % (
            config.DOMAIN_PROTOCOL,
            config.DOMAIN_NAME,
            playlist["id"],
            job["id"],
        )
        html = f"""<p>Hello {person.first_name},</p>
<p>Your playlist {playlist["name"]} build is available:
</p>
<p class="cta">
<a href="{playlist_url}">Download Playlist Build (.mp4)</a>
</p>
"""

        subject = f"{organisation['name']} - Kitsu: playlist download"
        title = "Playlist Download"
        email_html_body = templates_service.generate_html_body(title, html)
        emails.send_email(subject, email_html_body, email)


def get_playlist_download_context_name(project, playlist):
    """
    Build the context name (slug) for playlist download filenames.
    For tvshow projects, appends episode name (or "main pack" / "all assets").
    """
    context_name = slugify(project["name"], separator="_")
    if project.get("production_type") == "tvshow":
        episode_id = playlist.get("episode_id")
        if episode_id is not None:
            episode = shots_service.get_episode(episode_id)
            episode_name = episode["name"]
        elif playlist.get("is_for_all"):
            episode_name = "all assets"
        else:
            episode_name = "main pack"
        context_name += "_%s" % slugify(episode_name, separator="_")
    return context_name


def get_playlist_file_name(playlist):
    """
    Build file name for the movie file matching given playlist.
    """
    project = projects_service.get_project(playlist["project_id"])
    download_name = "%s_%s" % (
        slugify(project["name"]),
        slugify(playlist["name"]),
    )
    return slugify(download_name)


def get_playlist_movie_file_path(build_job):
    """
    Build file path for the movie file matching given playlist.
    """
    movie_file_name = "cache-playlists-%s.mp4" % build_job["id"]
    return os.path.join(config.TMP_DIR, movie_file_name)


def get_playlist_zip_file_path(playlist):
    """
    Build file path for the archive file matching given playlist.
    """
    zip_file_name = "%s.zip" % playlist["id"]
    return os.path.join(config.TMP_DIR, zip_file_name)


def get_build_job_raw(build_job_id):
    """
    Return given build job as active record.
    """
    return base_service.get_instance(
        BuildJob, build_job_id, BuildJobNotFoundException
    )


def get_build_job(build_job_id):
    """
    Return given build job as a dict.
    """
    return get_build_job_raw(build_job_id).serialize()


def remove_playlist(playlist_id):
    """
    Remove given playlist from database (and delete related build jobs).
    """
    playlist = get_playlist_raw(playlist_id)
    playlist_dict = playlist.serialize()
    jobs = BuildJob.query.filter_by(playlist_id=playlist_id).all()
    for job in jobs:
        _remove_build_job_impl(playlist_dict, job.serialize())
    playlist.delete()
    events.emit(
        "playlist:delete",
        {"playlist_id": playlist_dict["id"]},
        project_id=playlist_dict["project_id"],
    )
    return playlist_dict


def _remove_build_job_impl(playlist, job_dict):
    """
    Remove one build job (file + DB record + event). Caller must pass
    serialized job dict to avoid re-fetching when deleting multiple jobs.
    """
    build_job_id = job_dict["id"]
    movie_file_path = get_playlist_movie_file_path(job_dict)
    if os.path.exists(movie_file_path):
        os.remove(movie_file_path)
    if config.REMOVE_FILES:
        try:
            file_store.remove_movie("playlists", build_job_id)
        except Exception:
            current_app.logger.error(
                "Playlist file can't be deleted: %s" % build_job_id
            )
    job = BuildJob.get(build_job_id)
    if job is not None:
        job.delete()
    events.emit(
        "build-job:delete",
        {"build_job_id": build_job_id, "playlist_id": playlist["id"]},
        project_id=playlist["project_id"],
    )
    return movie_file_path


def remove_build_job(playlist, build_job_id):
    """
    Remove build job from database and remove related temporary file from
    hard drive.
    """
    job = BuildJob.get(build_job_id)
    if job is None:
        return None
    return _remove_build_job_impl(playlist, job.serialize())


def get_build_jobs_for_project(project_id):
    """
    Return all build_jobs for given project.
    """
    build_jobs = BuildJob.query.join(Playlist).filter(
        Playlist.project_id == project_id
    )
    return fields.serialize_list(build_jobs)


def get_playlists_for_project(project_id, page=0):
    """
    Return all playlists for given project.
    """
    query = Playlist.query.filter(Playlist.project_id == project_id)
    return query_utils.get_paginated_results(query, page, relations=True)


def generate_temp_playlist(task_ids, sort=True):
    """
    Generate the data structure of a playlist for a given task list. It doesn't
    persist anything. The goal is to build a temporary playlist used to see
    a quick preview of several shots.
    """
    if len(task_ids) == 0:
        return []

    task_id = task_ids[0]
    task = tasks_service.get_task(task_id)
    for_entity = entities_service.get_for_entity_from_task(task)
    task_type_links = projects_service.get_task_type_links(
        task["project_id"], for_entity
    )
    entities = []

    for task_id in task_ids:
        entity = generate_playlisted_entity_from_task(task_id, task_type_links)
        entities.append(entity)
    if len(entities) > 0:
        if not sort:
            return entities
        try:
            if "episode_name" in entities[0]:
                return sorted(entities, key=itemgetter("episode_name", "name"))
            elif "sequence_name" in entities[0]:
                return sorted(
                    entities, key=itemgetter("sequence_name", "name")
                )
            elif "asset_type_name" in entities[0]:
                return sorted(
                    entities, key=itemgetter("asset_type_name", "name")
                )
            else:
                return entities
        except Exception:
            return entities
    else:
        return []


def generate_playlisted_entity_from_task(task_id, task_type_links):
    """
    Generate the data structure of a playlisted shot for a given task. It
    doesn't persist anything.
    """
    task = tasks_service.get_task(task_id)
    entity = entities_service.get_entity(task["entity_id"])
    if shots_service.is_shot(entity):
        playlisted_entity = get_base_shot_for_playlist(entity, task_id)
    elif shots_service.is_sequence(entity):
        playlisted_entity = get_base_sequence_for_playlist(entity, task_id)
    elif shots_service.is_episode(entity):
        playlisted_entity = get_base_episode_for_playlist(entity, task_id)
    else:
        playlisted_entity = get_base_asset_for_playlist(entity, task_id)

    task_type_id = task["task_type_id"]
    preview_files = get_preview_files_for_entity(entity["id"])

    preview_file = None
    if task_type_id in preview_files and len(preview_files[task_type_id]) > 0:
        preview_file = preview_files[task_type_id][0]

    if preview_file is None and len(preview_files.keys()) > 0:
        available_task_types = [
            link["task_type_id"]
            for link in task_type_links
            if link["task_type_id"] in preview_files
        ]
        if len(available_task_types) > 0:
            task_type_id = available_task_types[0]
            preview_file = preview_files[task_type_id][0]

    if preview_file is not None:
        playlisted_entity.update(
            {
                "preview_file_id": preview_file["id"],
                "preview_file_extension": preview_file["extension"],
                "preview_file_width": preview_file["width"],
                "preview_file_height": preview_file["height"],
                "preview_file_duration": preview_file["duration"],
                "preview_file_revision": preview_file["revision"],
                "preview_file_status": preview_file["status"],
                "preview_file_annotations": preview_file["annotations"] or [],
                "preview_file_previews": preview_file["previews"],
            }
        )
    playlisted_entity["preview_files"] = preview_files
    return playlisted_entity


def get_base_episode_for_playlist(entity, task_id):
    episode = shots_service.get_episode(entity["id"])
    playlisted_entity = {
        "id": episode["id"],
        "name": episode["name"],
        "preview_file_task_id": task_id,
        "sequence_id": "",
        "sequence_name": "",
        "parent_name": "",
    }
    return playlisted_entity


def get_base_sequence_for_playlist(entity, task_id):
    sequence = shots_service.get_sequence(entity["id"])
    episode = None
    try:
        episode = shots_service.get_episode(sequence["parent_id"])
    except Exception as e:
        logger.debug(
            "No episode for sequence parent_id=%s: %s",
            sequence.get("parent_id"),
            e,
        )
    if episode is not None:
        playlisted_entity = {
            "id": sequence["id"],
            "name": sequence["name"],
            "preview_file_task_id": task_id,
            "episode_id": episode["id"],
            "episode_name": episode["name"],
            "parent_name": episode["name"],
        }
    else:
        playlisted_entity = {
            "id": sequence["id"],
            "name": sequence["name"],
            "preview_file_task_id": task_id,
            "sequence_id": "",
            "sequence_name": "",
            "parent_name": "",
        }
    return playlisted_entity


def get_base_shot_for_playlist(entity, task_id):
    shot = shots_service.get_shot(entity["id"])
    sequence = shots_service.get_sequence(shot["parent_id"])
    playlisted_entity = {
        "id": shot["id"],
        "name": shot["name"],
        "preview_file_task_id": task_id,
        "sequence_id": sequence["id"],
        "sequence_name": sequence["name"],
        "parent_name": sequence["name"],
    }
    return playlisted_entity


def get_base_asset_for_playlist(entity, task_id):
    asset = assets_service.get_asset(entity["id"])
    asset_type = assets_service.get_asset_type(asset["entity_type_id"])
    playlisted_entity = {
        "id": asset["id"],
        "name": asset["name"],
        "preview_file_task_id": task_id,
        "asset_type_id": asset_type["id"],
        "asset_type_name": asset_type["name"],
        "parent_name": asset_type["name"],
    }
    return playlisted_entity


def get_preview_files_for_task(task_id):
    """
    Return all preview file active records for given task.
    """
    preview_files = (
        PreviewFile.query.filter_by(task_id=task_id)
        .order_by(PreviewFile.revision.desc())
        .all()
    )
    return _get_playlist_preview_file_list(preview_files)


def _get_playlist_preview_file_list(preview_files):
    """
    Turn preview file active records into preview file dict that match the
    playlist data structure.
    """
    return [
        {
            "id": str(preview_file.id),
            "revision": preview_file.revision,
            "extension": preview_file.extension,
            "width": preview_file.width,
            "height": preview_file.height,
            "duration": float(preview_file.duration or 0),
            "status": str(preview_file.status),
            "annotations": preview_file.annotations,
            "created_at": fields.serialize_value(preview_file.created_at),
            "task_id": str(preview_file.task_id),
        }
        for preview_file in preview_files
    ]



from flask_restful import Resource

from flask_jwt_extended import jwt_required

from zou.app.mixin import ArgsMixin
from zou.app.utils import permissions
from zou.app.services import index_service, projects_service, user_service


class SearchResource(Resource, ArgsMixin):
    @jwt_required()
    def post(self):
        """
        Search entities
        ---
        description: Search across indexes for persons, assets and shots.
          Use optional filters to limit results to a project and specific
          indexes. Results are paginated with limit and offset.
        tags:
        - Search
        requestBody:
          required: true
          content:
            application/json:
              schema:
                type: object
                required:
                  - query
                properties:
                  query:
                    type: string
                    description: Search query string (minimum 3 characters)
                    example: "kitsu"
                  project_id:
                    type: string
                    format: uuid
                    description: Filter search results by project ID
                    example: a24a6ea4-ce75-4665-a070-57453082c25
                  limit:
                    type: integer
                    default: 3
                    description: Maximum number of results per index
                    example: 3
                  offset:
                    type: integer
                    default: 0
                    description: Number of results to skip
                    example: 0
                  index_names:
                    type: array
                    items:
                      type: string
                      enum: ["assets", "shots", "persons"]
                    default: ["assets", "shots", "persons"]
                    description: List of index names to search in
                    example: ["assets"]
        responses:
          200:
            description: List of entities that contain the query
            content:
              application/json:
                schema:
                  type: object
                  properties:
                    persons:
                      type: array
                      description: List of matching persons
                      example: [{
                          "id": "a24a6ea4-ce75-4665-a070-57453082c25",
                          "name": "John Doe",
                          ...
                        }]
                    assets:
                      type: array
                      description: List of matching assets
                      example: [{
                          "id": "a24a6ea4-ce75-4665-a070-57453082c25",
                          "name": "Chair prop CHR_001",
                          ...
                        }]
                      description: List of matching assets
                      example: []
                    shots:
                      type: array
                      description: List of matching shots
                      example: [{
                          "id": "a24a6ea4-ce75-4665-a070-57453082c25",
                          "name": "Shot 001",
                          ...
                        }]
          400:
            description: Bad request
        """
        args = self.get_args(
            [
                ("query", "", True),
                ("project_id", None, False),
                ("limit", 3, False, int),
                ("offset", 0, False, int),
                (
                    "index_names",
                    ["assets", "shots", "persons"],
                    False,
                    str,
                    "append",
                ),
            ]
        )
        query = args["query"]
        limit = args["limit"]
        offset = args["offset"]
        project_id = args["project_id"]
        index_names = args["index_names"]
        results = {}
        if len(query) < 3:
            return results

        if permissions.has_admin_permissions():
            project_ids = projects_service.open_project_ids()
        else:
            project_ids = user_service.get_open_project_ids()

        if project_id is not None and len(project_id) > 0:
            if project_id in project_ids:
                project_ids = [project_id]
            else:
                project_ids = []

        if "persons" in index_names:
            results["persons"] = index_service.search_persons(
                query, limit=limit, offset=offset
            )
        if "assets" in index_names:
            if (
                len(project_ids) == 0
                and not permissions.has_admin_permissions()
            ):
                results["assets"] = []
            else:
                results["assets"] = index_service.search_assets(
                    query, project_ids, limit=limit, offset=offset
                )
        if "shots" in index_names:
            if (
                len(project_ids) == 0
                and not permissions.has_admin_permissions()
            ):
                results["shots"] = []
            else:
                results["shots"] = index_service.search_shots(
                    query, project_ids, limit=limit, offset=offset
                )

        return results



from . import resources


routes = [(f"/hello-world", resources.HelloWorld)]


def pre_install(manifest):
    """
    Pre install the plugin.
    """


def post_install(manifest):
    """
    Post install the plugin.
    """


def pre_uninstall(manifest):
    """
    Pre uninstall the plugin.
    """


def post_uninstall(manifest):
    """
    Post uninstall the plugin.
    """



"""change_attachment_file_size_to_bigint

Revision ID: 6f4088b109fc
Revises: f45eed2a3ba6
Create Date: 2026-01-15 00:27:23.548198

"""

from alembic import op
import sqlalchemy as sa
import sqlalchemy_utils


# revision identifiers, used by Alembic.
revision = "6f4088b109fc"
down_revision = "f45eed2a3ba6"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("attachment_file", schema=None) as batch_op:
        batch_op.alter_column(
            "size",
            existing_type=sa.INTEGER(),
            type_=sa.BigInteger(),
            existing_nullable=True,
        )

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("attachment_file", schema=None) as batch_op:
        batch_op.alter_column(
            "size",
            existing_type=sa.BigInteger(),
            type_=sa.INTEGER(),
            existing_nullable=True,
        )

    # ### end Alembic commands ###



from gevent import monkey

monkey.patch_all()

from flask import jsonify
from flask_jwt_extended import (
    get_jwt_identity,
    jwt_required,
    verify_jwt_in_request,
)
from flask_socketio import SocketIO, disconnect, join_room, leave_room, emit

from zou.app import config, app
from zou.app.utils.redis import get_redis_url
from zou.app.services.playlists_service import get_playlist
from zou.app.services.user_service import check_project_access

server_stats = {"nb_connections": 0}
rooms_data = {}
user_rooms = {}

COMPARING_FIELDS = {
    "enable",
    "task_type",
    "revision",
    "mode",
    "comparison_preview_index",
}

redis_url = get_redis_url(config.KV_EVENTS_DB_INDEX)
socketio = SocketIO(
    logger=True, cors_allowed_origins=[], cors_credentials=False
)
socketio.init_app(app, message_queue=redis_url, async_mode="gevent")


def _check_room_access(playlist_id):
    """
    Check that the current user has access to the given playlist.
    Returns True if access is granted, False otherwise.
    """
    try:
        playlist = get_playlist(playlist_id)
        check_project_access(playlist["project_id"])
        return True
    except Exception:
        return False


def _add_user_to_room_index(user_id, room_id):
    user_rooms.setdefault(user_id, set()).add(room_id)


def _remove_user_from_room_index(user_id, room_id):
    if user_id in user_rooms:
        user_rooms[user_id].discard(room_id)
        if not user_rooms[user_id]:
            del user_rooms[user_id]


def _validate_comparing(data):
    """Sanitize the comparing dict to only keep known fields."""
    if not isinstance(data, dict):
        return None
    return {k: v for k, v in data.items() if k in COMPARING_FIELDS}


def _get_empty_room(current_frame=0):
    return {
        "playlist_id": None,
        "user_id": None,
        "local_id": None,
        "people": [],
        "is_playing": False,
        "current_entity_id": None,
        "current_entity_index": None,
        "current_preview_file_id": None,
        "current_preview_file_index": None,
        "current_frame": current_frame,
        "is_repeating": None,
        "is_annotations_displayed": False,
        "is_zoom_enabled": False,
        "is_waveform_displayed": False,
        "is_laser_mode": None,
        "handle_in": None,
        "handle_out": None,
        "speed": None,
        "comparing": {
            "enable": False,
            "task_type": None,
            "revision": None,
            "mode": "sidebyside",
            "comparison_preview_index": 0,
        },
    }


def _get_room_from_data(data):
    room_id = data.get("playlist_id", "0")
    return rooms_data.get(room_id, _get_empty_room()), room_id


def _leave_room(room_id, user_id):
    room = rooms_data.get(room_id, _get_empty_room())
    room["people"] = [p for p in room["people"] if p != user_id]
    _remove_user_from_room_index(user_id, room_id)
    if len(room["people"]) > 0:
        rooms_data[room_id] = room
    elif room_id in rooms_data:
        del rooms_data[room_id]
    _emit_people_updated(room_id, room["people"])
    return room


def _emit_people_updated(room_id, people):
    event_data = {
        "people": people,
        "playlist_id": room_id,
        "id": room_id,
    }
    emit("preview-room:room-people-updated", event_data, room=room_id)
    return event_data


def _update_room_playing_status(data, room):
    room["playlist_id"] = data.get("playlist_id", "")
    room["user_id"] = data.get("user_id", None)
    room["local_id"] = data.get("local_id", None)
    room["is_playing"] = data.get("is_playing", False)
    room["is_repeating"] = data.get("is_repeating", None)
    room["is_laser_mode"] = data.get("is_laser_mode", None)
    room["is_annotations_displayed"] = data.get(
        "is_annotations_displayed", False
    )
    room["is_zoom_enabled"] = data.get("is_zoom_enabled", False)
    room["is_waveform_displayed"] = data.get("is_waveform_displayed", False)
    room["current_entity_id"] = data.get("current_entity_id", None)
    room["current_entity_index"] = data.get("current_entity_index", None)
    room["current_preview_file_id"] = data.get("current_preview_file_id", None)
    room["current_preview_file_index"] = data.get(
        "current_preview_file_index", None
    )
    room["handle_in"] = data.get("handle_in", None)
    room["handle_out"] = data.get("handle_out", None)
    if "current_frame" in data:
        room["current_frame"] = data["current_frame"]
    if "comparing" in data:
        comparing = _validate_comparing(data["comparing"])
        if comparing is not None:
            room["comparing"] = comparing
    if "speed" in data and isinstance(data["speed"], (int, float)):
        room["speed"] = data["speed"]
    return room


@app.route("/", methods=["GET"])
def index():
    return jsonify({"name": "%s Event stream" % config.APP_NAME})


@app.route("/stats", methods=["GET"])
def stats():
    return jsonify(server_stats)


@socketio.on("connect", namespace="/events")
def connected(_):
    try:
        verify_jwt_in_request()
        server_stats["nb_connections"] += 1
        app.logger.info("New websocket client connected")
    except Exception:
        app.logger.info("New websocket client failed to connect")
        disconnect()
        return False


@socketio.on("disconnect", namespace="/events")
def disconnected(_):
    try:
        verify_jwt_in_request()
        user_id = get_jwt_identity()
        room_ids = list(user_rooms.get(user_id, set()))
        for room_id in room_ids:
            _leave_room(room_id, user_id)
            leave_room(room_id, user_id)
        app.logger.info("Websocket client disconnected")
    except Exception:
        app.logger.info("Websocket client disconnected (no valid token)")
    server_stats["nb_connections"] -= 1
    if server_stats["nb_connections"] < 0:
        server_stats["nb_connections"] = 0


@socketio.on_error("/events")
def on_error(error):
    server_stats["nb_connections"] -= 1
    if server_stats["nb_connections"] < 0:
        server_stats["nb_connections"] = 0
    app.logger.error(error)


@app.route("/rooms", methods=["GET", "POST"])
@jwt_required()
def rooms():
    return jsonify({"name": "%s Review rooms" % config.APP_NAME})


@socketio.on("preview-room:open-playlist", namespace="/events")
@jwt_required()
def on_open_playlist(data):
    """
    when a person opens the playlist page he immediately enters the
    websocket room. This way he can see in live which people are in the
    review room. The user still has to explicitly enter the review room
    to actually be in sync with the other users.
    """
    room, room_id = _get_room_from_data(data)
    if not _check_room_access(room_id):
        return
    rooms_data[room_id] = room
    join_room(room_id)
    _emit_people_updated(room_id, room["people"])


@socketio.on("preview-room:close-playlist", namespace="/events")
@jwt_required()
def on_close_playlist(data):
    """
    when a person closes the playlist page he immediately leaves the
    websocket room.
    """
    room_id = data.get("playlist_id", "0")
    leave_room(room_id)


@socketio.on("preview-room:join", namespace="/events")
@jwt_required()
def on_join(data):
    """
    When a person joins the review room, we notify all its members that a
    new person is added to the room.
    """
    user_id = get_jwt_identity()
    room, room_id = _get_room_from_data(data)
    if not _check_room_access(room_id):
        return
    if len(room["people"]) == 0:
        _update_room_playing_status(data, room)
    if user_id not in room["people"]:
        room["people"].append(user_id)
    room["playlist_id"] = room_id
    rooms_data[room_id] = room
    _add_user_to_room_index(user_id, room_id)
    _emit_people_updated(room_id, room["people"])
    emit("preview-room:room-updated", room, room=room_id)


@socketio.on("preview-room:leave", namespace="/events")
@jwt_required()
def on_leave(data):
    user_id = get_jwt_identity()
    room_id = data.get("playlist_id", "")
    _leave_room(room_id, user_id)


@socketio.on("preview-room:room-updated", namespace="/events")
@jwt_required()
def on_room_updated(data, only_newcomer=False):
    room, room_id = _get_room_from_data(data)
    rooms_data[room_id] = _update_room_playing_status(data, room)
    event_data = {"only_newcomer": only_newcomer, **rooms_data[room_id]}
    emit("preview-room:room-updated", event_data, room=room_id)


@socketio.on("preview-room:add-annotation", namespace="/events")
@jwt_required()
def on_add_annotation(data):
    room_id = data.get("playlist_id", "")
    emit("preview-room:add-annotation", data, room=room_id)


@socketio.on("preview-room:remove-annotation", namespace="/events")
@jwt_required()
def on_remove_annotation(data):
    room_id = data.get("playlist_id", "")
    emit("preview-room:remove-annotation", data, room=room_id)


@socketio.on("preview-room:update-annotation", namespace="/events")
@jwt_required()
def on_update_annotation(data):
    room_id = data.get("playlist_id", "")
    emit("preview-room:update-annotation", data, room=room_id)


@socketio.on("preview-room:change-version", namespace="/events")
@jwt_required()
def on_change_version(data):
    room_id = data.get("playlist_id", "")
    emit("preview-room:change-version", data, room=room_id)


@socketio.on("preview-room:panzoom-changed", namespace="/events")
@jwt_required()
def on_panzoom_changed(data):
    room_id = data.get("playlist_id", "")
    emit("preview-room:panzoom-changed", data, room=room_id)


@socketio.on("preview-room:comparison-panzoom-changed", namespace="/events")
@jwt_required()
def on_comparison_panzoom_changed(data):
    room_id = data.get("playlist_id", "")
    emit("preview-room:comparison-panzoom-changed", data, room=room_id)


if __name__ == "__main__":
    socketio.run(
        app,
        debug=False,
        host=config.EVENT_STREAM_HOST,
        port=config.EVENT_STREAM_PORT,
    )



from flask import current_app
from flask_jwt_extended import jwt_required
from sqlalchemy.exc import IntegrityError, StatementError

from zou.app.blueprints.crud.base import BaseModelsResource, BaseModelResource

from zou.app.models.status_automation import StatusAutomation
from zou.app.models.project import ProjectStatusAutomationLink
from zou.app.services import status_automations_service, user_service


class StatusAutomationsResource(BaseModelsResource):
    def __init__(self):
        BaseModelsResource.__init__(self, StatusAutomation)

    def check_read_permissions(self, options=None):
        user_service.block_access_to_vendor()
        return True

    @jwt_required()
    def get(self):
        """
        Get status automations
        ---
        tags:
          - Crud
        description: Retrieve all status automations. Supports filtering
          via query parameters and pagination. Vendor access is blocked.
        parameters:
          - in: query
            name: page
            required: false
            schema:
              type: integer
            example: 1
            description: Page number for pagination
          - in: query
            name: limit
            required: false
            schema:
              type: integer
            example: 50
            description: Number of results per page
          - in: query
            name: relations
            required: false
            schema:
              type: boolean
            default: false
            example: false
            description: Whether to include relations
        responses:
            200:
              description: Status automations retrieved successfully
              content:
                application/json:
                  schema:
                    oneOf:
                      - type: array
                        items:
                          type: object
                      - type: object
                        properties:
                          data:
                            type: array
                            items:
                              type: object
                            example: []
                          total:
                            type: integer
                            example: 100
                          nb_pages:
                            type: integer
                            example: 2
                          limit:
                            type: integer
                            example: 50
                          offset:
                            type: integer
                            example: 0
                          page:
                            type: integer
                            example: 1
            400:
              description: Invalid filter format or query error
        """
        return super().get()

    @jwt_required()
    def post(self):
        """
        Create status automation
        ---
        tags:
          - Crud
        description: Create a new status automation with data provided
          in the request body. JSON format is expected.
        requestBody:
          required: true
          content:
            application/json:
              schema:
                type: object
                required:
                  - source_task_status_id
                  - target_task_status_id
                  - field_name
                  - field_value
                properties:
                  source_task_status_id:
                    type: string
                    format: uuid
                    example: a24a6ea4-ce75-4665-a070-57453082c25
                  target_task_status_id:
                    type: string
                    format: uuid
                    example: b24a6ea4-ce75-4665-a070-57453082c25
                  field_name:
                    type: string
                    example: ready_for
                  field_value:
                    type: string
                    example: production
        responses:
            201:
              description: Status automation created successfully
              content:
                application/json:
                  schema:
                    type: object
                    properties:
                      id:
                        type: string
                        format: uuid
                        example: a24a6ea4-ce75-4665-a070-57453082c25
                      source_task_status_id:
                        type: string
                        format: uuid
                        example: b24a6ea4-ce75-4665-a070-57453082c25
                      target_task_status_id:
                        type: string
                        format: uuid
                        example: c24a6ea4-ce75-4665-a070-57453082c25
                      field_name:
                        type: string
                        example: ready_for
                      field_value:
                        type: string
                        example: production
                      created_at:
                        type: string
                        format: date-time
                        example: "2024-01-15T10:30:00Z"
                      updated_at:
                        type: string
                        format: date-time
                        example: "2024-01-15T10:30:00Z"
            400:
              description: Invalid data format or validation error
        """
        return super().post()

    def post_creation(self, status_automation):
        status_automations_service.clear_status_automation_cache()
        return status_automation.serialize()


class StatusAutomationResource(BaseModelResource):
    def __init__(self):
        BaseModelResource.__init__(self, StatusAutomation)

    @jwt_required()
    def get(self, instance_id):
        """
        Get status automation
        ---
        tags:
          - Crud
        description: Retrieve a status automation by its ID and return
          it as a JSON object. Supports including relations.
        parameters:
          - in: path
            name: instance_id
            required: true
            schema:
              type: string
              format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
          - in: query
            name: relations
            required: false
            schema:
              type: boolean
            default: true
            example: true
            description: Whether to include relations
        responses:
            200:
              description: Status automation retrieved successfully
              content:
                application/json:
                  schema:
                    type: object
                    properties:
                      id:
                        type: string
                        format: uuid
                        example: a24a6ea4-ce75-4665-a070-57453082c25
                      source_task_status_id:
                        type: string
                        format: uuid
                        example: b24a6ea4-ce75-4665-a070-57453082c25
                      target_task_status_id:
                        type: string
                        format: uuid
                        example: c24a6ea4-ce75-4665-a070-57453082c25
                      field_name:
                        type: string
                        example: ready_for
                      field_value:
                        type: string
                        example: production
                      created_at:
                        type: string
                        format: date-time
                        example: "2024-01-15T10:30:00Z"
                      updated_at:
                        type: string
                        format: date-time
                        example: "2024-01-15T10:30:00Z"
            400:
              description: Invalid ID format or query error
        """
        return super().get(instance_id)

    @jwt_required()
    def put(self, instance_id):
        """
        Update status automation
        ---
        tags:
          - Crud
        description: Update a status automation with data provided in
          the request body. JSON format is expected.
        parameters:
          - in: path
            name: instance_id
            required: true
            schema:
              type: string
              format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
        requestBody:
          required: true
          content:
            application/json:
              schema:
                type: object
                properties:
                  field_name:
                    type: string
                    example: ready_for
                  field_value:
                    type: string
                    example: approved
        responses:
            200:
              description: Status automation updated successfully
              content:
                application/json:
                  schema:
                    type: object
                    properties:
                      id:
                        type: string
                        format: uuid
                        example: a24a6ea4-ce75-4665-a070-57453082c25
                      source_task_status_id:
                        type: string
                        format: uuid
                        example: b24a6ea4-ce75-4665-a070-57453082c25
                      target_task_status_id:
                        type: string
                        format: uuid
                        example: c24a6ea4-ce75-4665-a070-57453082c25
                      field_name:
                        type: string
                        example: ready_for
                      field_value:
                        type: string
                        example: approved
                      created_at:
                        type: string
                        format: date-time
                        example: "2024-01-15T10:30:00Z"
                      updated_at:
                        type: string
                        format: date-time
                        example: "2024-01-15T11:00:00Z"
            400:
              description: Invalid data format or validation error
        """
        return super().put(instance_id)

    def post_update(self, status_automation, data):
        status_automations_service.clear_status_automation_cache()
        return status_automation

    @jwt_required()
    def delete(self, instance_id):
        """
        Delete status automation
        ---
        tags:
          - Crud
        description: Delete a status automation by its ID. Returns empty
          response on success. Cannot delete if automation is used in a
          project.
        parameters:
          - in: path
            name: instance_id
            required: true
            schema:
              type: string
              format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
        responses:
            204:
              description: Status automation deleted successfully
            400:
              description: Automation is used in a project or integrity error
        """
        instance = self.get_model_or_404(instance_id)
        instance_dict = instance.serialize()
        links = ProjectStatusAutomationLink.query.filter_by(
            status_automation_id=instance_dict["id"]
        ).all()
        if len(links) > 0:
            return {"message": "This automation is used in a project."}, 400

        try:
            self.check_delete_permissions(instance_dict)
            self.pre_delete(instance_dict)
            instance.delete()
            self.emit_delete_event(instance_dict)
            self.post_delete(instance_dict)

        except IntegrityError as exception:
            current_app.logger.error(str(exception), exc_info=1)
            return {"message": str(exception)}, 400

        except StatementError as exception:
            current_app.logger.error(str(exception), exc_info=1)
            return {"message": str(exception)}, 400

        return "", 204

    def post_delete(self, status_automation):
        status_automations_service.clear_status_automation_cache()
        return status_automation



import slugify

from sqlalchemy.exc import IntegrityError

from zou.app.models.preview_background_file import PreviewBackgroundFile
from zou.app.models.entity import Entity
from zou.app.models.entity_type import EntityType
from zou.app.models.metadata_descriptor import MetadataDescriptor
from zou.app.models.person import Person, DepartmentLink
from zou.app.models.project import (
    Project,
    ProjectPersonLink,
    ProjectTaskTypeLink,
    ProjectTaskStatusLink,
)
from zou.app.models.project_status import ProjectStatus
from zou.app.models.status_automation import StatusAutomation
from zou.app.models.task_type import TaskType
from zou.app.models.task_status import TaskStatus
from zou.app.models.department import Department
from zou.app.services import (
    assets_service,
    base_service,
    edits_service,
    shots_service,
)
from zou.app.services.exception import (
    ProjectNotFoundException,
    MetadataDescriptorNotFoundException,
    DepartmentNotFoundException,
    WrongParameterException,
)

from zou.app.utils import fields, events, cache

from sqlalchemy.exc import StatementError
from sqlalchemy.orm import joinedload
from sqlalchemy.orm.exc import ObjectDeletedError
from sqlalchemy import or_


def clear_project_cache(project_id):
    cache.cache.delete_memoized(get_project, project_id)
    cache.cache.delete_memoized(get_project, project_id, True)
    cache.cache.delete_memoized(get_project_by_name)
    cache.cache.delete_memoized(open_projects)


@cache.memoize_function(120)
def open_projects(name=None):
    """
    Return all open projects. Allow to filter projects by name.
    """
    query = (
        Project.query.join(
            ProjectStatus, Project.project_status_id == ProjectStatus.id
        )
        .outerjoin(MetadataDescriptor)
        .filter(ProjectStatus.name.in_(("Active", "open", "Open")))
        .order_by(Project.name)
    )

    if name is not None:
        query = query.filter(Project.name == name)

    return get_projects_with_extra_data(query)


def open_project_ids():
    """
    Return all open project ids. Allow to filter projects by name.
    """
    return [project["id"] for project in open_projects()]


def get_projects_with_extra_data(
    query, for_client=False, vendor_departments=None
):
    """
    Helpers function to attach:
    * First episode name to current project when it's a TV Show.
    * Add metadata descriptors for this project.
    * Add task types and task statuses for this project.
    """
    projects_list = query.all()
    if not projects_list:
        return []

    project_ids = [p.id for p in projects_list]

    descriptors_by_project = _fetch_metadata_descriptors_by_project(
        project_ids, for_client, vendor_departments
    )
    task_types_by_project = _fetch_task_type_links_by_project(project_ids)
    task_statuses_by_project = _fetch_task_status_links_by_project(project_ids)
    tvshow_project_ids = [
        p.id for p in projects_list if p.production_type == "tvshow"
    ]
    first_episodes_by_project = _fetch_first_episodes_by_project(
        tvshow_project_ids
    )
    return [
        _build_project_dict_with_extra_data(
            project,
            descriptors_by_project,
            task_types_by_project,
            task_statuses_by_project,
            first_episodes_by_project,
        )
        for project in projects_list
    ]


def _fetch_metadata_descriptors_by_project(
    project_ids, for_client=False, vendor_departments=None
):
    if for_client:
        descriptors_query = MetadataDescriptor.query.filter(
            MetadataDescriptor.project_id.in_(project_ids),
            MetadataDescriptor.for_client == True,
        )
    elif vendor_departments is not None:
        descriptors_query = MetadataDescriptor.query.filter(
            MetadataDescriptor.project_id.in_(project_ids)
        ).filter(
            or_(
                MetadataDescriptor.departments == None,
                MetadataDescriptor.departments.any(
                    Department.id.in_(vendor_departments)
                ),
            )
        )
    else:
        descriptors_query = MetadataDescriptor.query.filter(
            MetadataDescriptor.project_id.in_(project_ids)
        )
    # Eager-load departments to avoid N+1 when serializing descriptors
    descriptors_query = descriptors_query.options(
        joinedload(MetadataDescriptor.departments)
    )

    all_descriptors = descriptors_query.all()
    descriptors_by_project = {}
    for desc in all_descriptors:
        if desc.project_id not in descriptors_by_project:
            descriptors_by_project[desc.project_id] = []
        descriptors_by_project[desc.project_id].append(desc)
    return descriptors_by_project


def _fetch_task_type_links_by_project(project_ids):
    task_type_links = ProjectTaskTypeLink.query.filter(
        ProjectTaskTypeLink.project_id.in_(project_ids)
    ).all()
    task_types_by_project = {}
    for link in task_type_links:
        if link.project_id not in task_types_by_project:
            task_types_by_project[link.project_id] = {}
        task_types_by_project[link.project_id][
            str(link.task_type_id)
        ] = link.priority
    return task_types_by_project


def _fetch_task_status_links_by_project(project_ids):
    task_status_links = ProjectTaskStatusLink.query.filter(
        ProjectTaskStatusLink.project_id.in_(project_ids)
    ).all()
    task_statuses_by_project = {}
    for link in task_status_links:
        if link.project_id not in task_statuses_by_project:
            task_statuses_by_project[link.project_id] = {}
        task_statuses_by_project[link.project_id][str(link.task_status_id)] = {
            "priority": link.priority,
            "roles_for_board": fields.serialize_list(link.roles_for_board),
        }
    return task_statuses_by_project


def _fetch_first_episodes_by_project(project_ids):
    if not project_ids:
        return {}

    episode_type = shots_service.get_episode_type()
    first_episodes_by_project = {}

    episodes = (
        Entity.query.filter(
            Entity.project_id.in_(project_ids),
            Entity.entity_type_id == episode_type["id"],
            Entity.status == "running",
        )
        .order_by(Entity.project_id, Entity.name)
        .all()
    )

    seen_projects = set()
    for episode in episodes:
        if episode.project_id not in seen_projects:
            first_episodes_by_project[episode.project_id] = episode
            seen_projects.add(episode.project_id)

    missing_projects = set(project_ids) - seen_projects
    if missing_projects:
        fallback_episodes = (
            Entity.query.filter(
                Entity.project_id.in_(missing_projects),
                Entity.entity_type_id == episode_type["id"],
            )
            .order_by(Entity.project_id, Entity.name)
            .all()
        )

        for episode in fallback_episodes:
            if episode.project_id not in first_episodes_by_project:
                first_episodes_by_project[episode.project_id] = episode

    return first_episodes_by_project


def _serialize_descriptor(descriptor):
    return {
        "id": fields.serialize_value(descriptor.id),
        "name": descriptor.name,
        "field_name": descriptor.field_name,
        "data_type": fields.serialize_value(descriptor.data_type),
        "choices": descriptor.choices,
        "for_client": descriptor.for_client or False,
        "entity_type": descriptor.entity_type,
        "position": descriptor.position,
        "departments": [
            str(department.id) for department in descriptor.departments
        ],
    }


def _build_project_dict_with_extra_data(
    project,
    descriptors_by_project,
    task_types_by_project,
    task_statuses_by_project,
    first_episodes_by_project,
):
    project_dict = project.serialize(relations=True)

    project_dict["descriptors"] = [
        _serialize_descriptor(descriptor)
        for descriptor in descriptors_by_project.get(project.id, [])
    ]
    project_dict["task_types_priority"] = task_types_by_project.get(
        project.id, {}
    )
    project_dict["task_statuses_link"] = task_statuses_by_project.get(
        project.id, {}
    )
    if project.production_type == "tvshow":
        episode = first_episodes_by_project.get(project.id)
        if episode is not None:
            project_dict["first_episode_id"] = fields.serialize_value(
                episode.id
            )

    return project_dict


def get_projects():
    """
    Return all projects. Allow to filter projects by name.
    """
    query = (
        Project.query.join(
            ProjectStatus, Project.project_status_id == ProjectStatus.id
        )
        .add_columns(ProjectStatus.name)
        .order_by(Project.name)
    )

    result = []
    for entry in query.all():
        (project, project_status_name) = entry
        data = project.serialize()
        data["project_status_name"] = project_status_name
        result.append(data)

    return result


@cache.memoize_function(480)
def get_project_statuses():
    return fields.serialize_models(ProjectStatus.get_all())


def get_or_create_open_status():
    """
    Return open status. If it does not exist, it creates it.
    """
    return get_or_create_status("Open")


@cache.memoize_function(480)
def get_open_status():
    """
    Return open status. If it does not exist, it creates it.
    """
    return get_or_create_status("Open")


@cache.memoize_function(120)
def get_closed_status():
    """
    Return closed status. If it does not exist, it creates it.
    """
    return get_or_create_status("Closed")


def get_or_create_status(name):
    """
    Return given status. If it does not exist, it creates it.
    """
    project_status = ProjectStatus.get_by(name=name)
    if project_status is None:
        project_status = ProjectStatus(name=name, color="#000000")
        project_status.save()
    return project_status.serialize()


def save_project_status(project_statuses):
    """
    Save in database all project status given in parameter.
    """
    result = []
    filtered_satuses = (x for x in project_statuses if x is not None)

    for status in filtered_satuses:
        project_status = get_or_create_status(status)
        result.append(project_status)
    return result


def get_or_create_project(name):
    """
    Get project which match given name. Create it if it does not exist.
    """
    project = Project.get_by(name=name)
    if project is None:
        open_status = get_or_create_open_status()
        project = Project(name=name, project_status_id=open_status["id"])
        project.save()
    return project.serialize()


def get_project_raw(project_id):
    """
    Get project matching given id, as active record. Raises an exception if
    project is not found.
    """
    try:
        project = Project.get(project_id)
    except StatementError:
        raise ProjectNotFoundException()

    if project is None:
        raise ProjectNotFoundException()

    return project


@cache.memoize_function(240)
def get_project(project_id, relations=False):
    """
    Get project matching given id, as a dict. Raises an exception if project is
    not found.
    """
    return get_project_raw(project_id).serialize(relations=relations)


@cache.memoize_function(120)
def get_project_by_name(project_name):
    """
    Get project matching given name. Raises an exception if project is not
    found.
    """
    project = Project.query.filter(Project.name.ilike(project_name)).first()

    if project is None:
        raise ProjectNotFoundException()

    return project.serialize()


def update_project(project_id, data):
    """
    Update project matching given id with data from *data* dict.
    """
    project = get_project_raw(project_id)
    project.update(data)
    clear_project_cache(project_id)
    events.emit("project:update", {}, project_id=project_id)
    return project.serialize()


def add_team_member(project_id, person_id):
    """
    Add a person listed in database to the the project team.
    """
    return _add_to_list_attr(project_id, Person, person_id, "team")


def remove_team_member(project_id, person_id):
    """
    Remove a person listed in database from the the project team.
    """
    return _remove_from_list_attr(project_id, Person, person_id, "team")


def add_asset_type_setting(project_id, asset_type_id):
    """
    Add an asset type listed in database to the the project asset types.
    """
    return _add_to_list_attr(
        project_id, EntityType, asset_type_id, "asset_types"
    )


def remove_asset_type_setting(project_id, asset_type_id):
    """
    Remove an asset type listed in database from the the project asset types.
    """
    return _remove_from_list_attr(
        project_id, EntityType, asset_type_id, "asset_types"
    )


def add_task_type_setting(project_id, task_type_id, priority=None):
    """
    Add a task type listed in database to the the project task types.
    """
    project_id = str(project_id)
    task_type_id = str(task_type_id)
    if not task_type_id or not fields.is_valid_id(task_type_id):
        raise WrongParameterException(
            "task_type_id is required and must be a valid UUID"
        )

    link = ProjectTaskTypeLink.get_by(
        task_type_id=task_type_id, project_id=project_id
    )
    if not link:
        ProjectTaskTypeLink.create(
            task_type_id=task_type_id, project_id=project_id, priority=priority
        )
    return _save_project(get_project_raw(project_id))


def remove_task_type_setting(project_id, task_type_id):
    """
    Remove a task status listed in database from the the project task types.
    """
    return _remove_from_list_attr(
        project_id, TaskType, task_type_id, "task_types"
    )


def add_task_status_setting(project_id, task_status_id):
    """
    Add a task status listed in database to the the project task statuses.
    """
    return _add_to_list_attr(
        project_id, TaskStatus, task_status_id, "task_statuses"
    )


def remove_task_status_setting(project_id, task_status_id):
    """
    Remove a task status listed in database from the the project task statuses.
    """
    return _remove_from_list_attr(
        project_id, TaskStatus, task_status_id, "task_statuses"
    )


def add_status_automation_setting(project_id, status_automation_id):
    """
    Add a status automation listed in database to the project status automations.
    """
    return _add_to_list_attr(
        project_id,
        StatusAutomation,
        status_automation_id,
        "status_automations",
    )


def remove_status_automation_setting(project_id, status_automation_id):
    """
    Remove a status automation listed in database from the project status
    automations.
    """
    return _remove_from_list_attr(
        project_id,
        StatusAutomation,
        status_automation_id,
        "status_automations",
    )


def add_preview_background_file_setting(
    project_id, preview_background_file_id
):
    """
    Add a preview background file listed in database to the project backgrounds.
    """
    return _add_to_list_attr(
        project_id,
        PreviewBackgroundFile,
        preview_background_file_id,
        "preview_background_files",
    )


def remove_preview_background_file_setting(
    project_id, preview_background_file_id
):
    """
    Remove a preview background file listed in database from the project backgrounds.
    """
    return _remove_from_list_attr(
        project_id,
        PreviewBackgroundFile,
        preview_background_file_id,
        "preview_background_files",
    )


def _add_to_list_attr(project_id, model_class, model_id, list_attr):
    project = get_project_raw(project_id)
    model = model_class.get(model_id)
    if model is None:
        model_name = model_class.__name__
        raise WrongParameterException(
            f"{model_name} with id {model_id} not found"
        )
    if str(model.id) not in [str(m.id) for m in getattr(project, list_attr)]:
        getattr(project, list_attr).append(model)
        return _save_project(project)
    else:
        return project.serialize()


def _remove_from_list_attr(project_id, model_class, model_id, list_attr):
    project = get_project_raw(project_id)
    model = model_class.get(model_id)
    try:
        getattr(project, list_attr).remove(model)
    except ValueError:
        pass
    return _save_project(project)


def _save_project(project):
    project.save()
    clear_project_cache(str(project.id))
    events.emit("project:update", {}, project_id=str(project.id))
    return project.serialize()


def add_metadata_descriptor(
    project_id,
    entity_type,
    name,
    data_type,
    choices,
    for_client,
    departments=None,
):
    if not departments:
        departments = []

    try:
        departments_objects = [
            Department.get(department_id)
            for department_id in departments
            if department_id is not None
        ]
    except StatementError:
        raise DepartmentNotFoundException()

    try:
        descriptor = MetadataDescriptor.create(
            project_id=project_id,
            entity_type=entity_type,
            name=name,
            data_type=data_type,
            choices=choices,
            for_client=for_client,
            departments=departments_objects,
            field_name=slugify.slugify(name, separator="_"),
        )
    except IntegrityError:
        raise WrongParameterException("Metadata descriptor already exists.")
    events.emit(
        "metadata-descriptor:new",
        {"metadata_descriptor_id": str(descriptor.id)},
        project_id=project_id,
    )
    clear_project_cache(project_id)
    return descriptor.serialize()


def get_metadata_descriptors(project_id, for_client=False):
    """
    Get all metadata descriptors for given project and entity type.
    """
    query = MetadataDescriptor.query.filter(
        MetadataDescriptor.project_id == project_id
    ).order_by(MetadataDescriptor.position, MetadataDescriptor.name)
    if for_client:
        query = query.filter(MetadataDescriptor.for_client == True)

    # Eager-load departments to avoid N+1 during serialization
    query = query.options(joinedload(MetadataDescriptor.departments))
    descriptors = query.all()
    return fields.serialize_models(descriptors, relations=True)


def get_metadata_descriptor_raw(metadata_descriptor_id):
    """
    Get metadata descriptor for given id as active record.
    """
    return base_service.get_instance(
        MetadataDescriptor,
        metadata_descriptor_id,
        MetadataDescriptorNotFoundException,
    )


def get_metadata_descriptor(metadata_descriptor_id):
    """
    Get metadata descriptor for given id as dict.
    """
    return get_metadata_descriptor_raw(metadata_descriptor_id).serialize(
        relations=True
    )


def update_metadata_descriptor(metadata_descriptor_id, changes):
    """
    Update metadata descriptor information for given id.
    """
    descriptor = get_metadata_descriptor_raw(metadata_descriptor_id)

    if "name" in changes and len(changes["name"]) > 0:
        changes["field_name"] = slugify.slugify(changes["name"], separator="_")
        if descriptor.field_name != changes["field_name"]:
            query = Entity.query.filter(
                Entity.project_id == descriptor.project_id
            )
            if descriptor.entity_type == "Shot":
                shot_type = shots_service.get_shot_type()
                query = query.filter(Entity.entity_type_id == shot_type["id"])
            elif descriptor.entity_type == "Asset":
                query = query.filter(assets_service.build_asset_type_filter())
            elif descriptor.entity_type == "Edit":
                edit_type = edits_service.get_edit_type()
                query = query.filter(Entity.entity_type_id == edit_type["id"])

            entities = query.all()
            for entity in entities:
                metadata = fields.serialize_value(entity.data) or {}
                value = metadata.pop(descriptor.field_name, None)
                if value is not None:
                    metadata[changes["field_name"]] = value
                    entity.update_no_commit({"data": metadata})
            Entity.commit()

    if "departments" in changes:
        if not changes["departments"]:
            changes["departments"] = []

        try:
            departments_objects = [
                Department.get(department_id)
                for department_id in changes["departments"]
                if department_id is not None
            ]
        except StatementError:
            raise DepartmentNotFoundException()

        changes["departments"] = departments_objects

    descriptor.update(changes)
    events.emit(
        "metadata-descriptor:update",
        {"metadata_descriptor_id": str(descriptor.id)},
        project_id=descriptor.project_id,
    )
    clear_project_cache(str(descriptor.project_id))
    return descriptor.serialize()


def reorder_metadata_descriptors(project_id, entity_type, descriptor_ids):
    """
    Reorder metadata descriptors for a given project and entity type.
    Updates position field based on the order of descriptor IDs provided.
    Descriptors not in the list are added at the end, ordered by name.
    """
    descriptors = MetadataDescriptor.query.filter(
        MetadataDescriptor.project_id == project_id,
        MetadataDescriptor.entity_type == entity_type,
    ).all()

    descriptor_map = {str(desc.id): desc for desc in descriptors}

    for descriptor_id in descriptor_ids:
        if descriptor_id not in descriptor_map:
            raise WrongParameterException(
                f"Descriptor {descriptor_id} not found for project {project_id} and entity type {entity_type}"
            )

    for position, descriptor_id in enumerate(descriptor_ids, start=1):
        descriptor = descriptor_map[descriptor_id]
        descriptor.update({"position": position})

    descriptors_not_in_list = [
        desc for desc in descriptors if not str(desc.id) in descriptor_ids
    ]
    descriptors_not_in_list.sort(key=lambda d: d.name)
    start_position = len(descriptor_ids) + 1
    for position_offset, descriptor in enumerate(descriptors_not_in_list):
        descriptor.update({"position": start_position + position_offset})

    clear_project_cache(project_id)

    query = MetadataDescriptor.query.filter(
        MetadataDescriptor.project_id == project_id,
        MetadataDescriptor.entity_type == entity_type,
    ).order_by(MetadataDescriptor.position, MetadataDescriptor.name)
    # Eager-load departments to avoid N+1 during serialization
    query = query.options(joinedload(MetadataDescriptor.departments))
    return fields.serialize_models(query.all(), relations=True)


def remove_metadata_descriptor(metadata_descriptor_id):
    """
    Delete metadata descriptor and related informations.
    """
    descriptor = get_metadata_descriptor_raw(metadata_descriptor_id)
    entities = Entity.get_all_by(project_id=descriptor.project_id)
    for entity in entities:
        metadata = fields.serialize_value(entity.data)
        if metadata is not None:
            metadata.pop(descriptor.field_name, None)
            entity.update({"data": metadata})
    try:
        descriptor.delete()
    except ObjectDeletedError:
        pass
    events.emit(
        "metadata-descriptor:delete",
        {"metadata_descriptor_id": str(descriptor.id)},
        project_id=descriptor.project_id,
    )
    clear_project_cache(str(descriptor.project_id))
    return descriptor.serialize()


def is_tv_show(project):
    return project["production_type"] == "tvshow"


def is_open(project):
    open_status = get_open_status()
    return project["project_status_id"] == open_status["id"]


def create_project_task_type_link(project_id, task_type_id, priority):
    if not task_type_id or not fields.is_valid_id(task_type_id):
        raise WrongParameterException(
            "task_type_id is required and must be a valid UUID"
        )

    task_type_link = ProjectTaskTypeLink.get_by(
        project_id=project_id, task_type_id=task_type_id
    )

    if priority is not None:
        priority = int(priority)

    if task_type_link is None:
        task_type_link = ProjectTaskTypeLink.create(
            project_id=project_id, task_type_id=task_type_id, priority=priority
        )
    else:
        task_type_link.update({"priority": priority})

    return task_type_link.serialize()


def create_project_task_status_link(
    project_id, task_status_id, priority, roles_for_board=None
):
    if not task_status_id or not fields.is_valid_id(task_status_id):
        raise WrongParameterException(
            "task_status_id is required and must be a valid UUID"
        )

    task_status_link = ProjectTaskStatusLink.get_by(
        project_id=project_id, task_status_id=task_status_id
    )

    if priority is not None:
        priority = int(priority)

    if task_status_link is None:
        task_status_link = ProjectTaskStatusLink.create(
            project_id=project_id,
            task_status_id=task_status_id,
            priority=priority,
            roles_for_board=roles_for_board,
        )
    else:
        task_status_link.update(
            {"priority": priority, "roles_for_board": roles_for_board}
        )

    return task_status_link.serialize()


def get_project_task_types(project_id):
    project = get_project_raw(project_id)
    return Project.serialize_list(project.task_types)


def get_project_task_statuses(project_id):
    project = get_project_raw(project_id)
    return Project.serialize_list(project.task_statuses)


def get_project_status_automations(project_id):
    project = get_project_raw(project_id)
    return Project.serialize_list(project.status_automations)


def get_project_preview_background_files(project_id):
    project = get_project_raw(project_id)
    return Project.serialize_list(project.preview_background_files)


def get_project_fps(project_id):
    """
    Return fps set at project level or default fps if it not set.
    """
    project = get_project(project_id)
    return float(project["fps"] or "25.00")


def get_task_type_priority_map(project_id, for_entity="Asset"):
    """
    Return a dict allowing to match a task type id with a priority.
    """
    task_types = (
        ProjectTaskTypeLink.query.join(TaskType)
        .filter(ProjectTaskTypeLink.project_id == project_id)
        .filter(TaskType.for_entity == for_entity)
    ).all()
    return {
        str(task_type_link.task_type_id): task_type_link.priority
        for task_type_link in task_types
    }


def get_task_type_links(project_id, for_entity="Asset"):
    """
    Return a lisk of links for given project and entity type.
    """
    task_type_links = (
        ProjectTaskTypeLink.query.join(TaskType)
        .filter(ProjectTaskTypeLink.project_id == project_id)
        .filter(TaskType.for_entity == for_entity)
        .order_by(ProjectTaskTypeLink.priority.desc())
    ).all()
    return ProjectTaskTypeLink.serialize_list(task_type_links)


def get_department_team(project_id, department_id):
    """
    Get all persons in a given department for a given project.
    """
    persons = (
        Person.query.join(
            ProjectPersonLink, ProjectPersonLink.person_id == Person.id
        )
        .join(DepartmentLink, DepartmentLink.person_id == Person.id)
        .filter(ProjectPersonLink.project_id == project_id)
        .filter(DepartmentLink.department_id == department_id)
    ).all()
    return persons



"""add indexes for better performance

Revision ID: 68bae7fcd569
Revises: 6f4088b109fc
Create Date: 2026-01-15 23:33:33.926956

"""

from alembic import op
import sqlalchemy as sa
import sqlalchemy_utils


# revision identifiers, used by Alembic.
revision = "68bae7fcd569"
down_revision = "6f4088b109fc"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("entity", schema=None) as batch_op:
        batch_op.create_index(
            "ix_entity_project_type_status",
            ["project_id", "entity_type_id", "status"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_entity_status"), ["status"], unique=False
        )

    with op.batch_alter_table(
        "project_asset_type_link", schema=None
    ) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_project_asset_type_link_asset_type_id"),
            ["asset_type_id"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_project_asset_type_link_project_id"),
            ["project_id"],
            unique=False,
        )

    with op.batch_alter_table(
        "project_preview_background_file_link", schema=None
    ) as batch_op:
        batch_op.create_index(
            batch_op.f(
                "ix_project_preview_background_file_link_preview_background_file_id"
            ),
            ["preview_background_file_id"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_project_preview_background_file_link_project_id"),
            ["project_id"],
            unique=False,
        )

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table(
        "project_preview_background_file_link", schema=None
    ) as batch_op:
        batch_op.drop_index(
            batch_op.f("ix_project_preview_background_file_link_project_id")
        )
        batch_op.drop_index(
            batch_op.f(
                "ix_project_preview_background_file_link_preview_background_file_id"
            )
        )

    with op.batch_alter_table(
        "project_asset_type_link", schema=None
    ) as batch_op:
        batch_op.drop_index(
            batch_op.f("ix_project_asset_type_link_project_id")
        )
        batch_op.drop_index(
            batch_op.f("ix_project_asset_type_link_asset_type_id")
        )

    with op.batch_alter_table("entity", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_entity_status"))
        batch_op.drop_index("ix_entity_project_type_status")

    # ### end Alembic commands ###



import orjson as json

from flask import abort, request, send_file as flask_send_file, current_app
from flask_restful import Resource, reqparse
from flask_jwt_extended import jwt_required

from zou.app.mixin import ArgsMixin
from zou.app.utils import permissions, date_helpers

from zou.app.services import (
    chats_service,
    comments_service,
    deletion_service,
    entities_service,
    persons_service,
    tasks_service,
    user_service,
)
from zou.app import config


class DownloadAttachmentResource(Resource):

    @jwt_required()
    def get(self, attachment_file_id, file_name):
        """
        Download attachment file
        ---
        description: Download a specific attachment file from a comment or chat
          message. Supports various file types including images and documents.
        tags:
          - Comments
        produces:
          - multipart/form-data
          - image/png
          - image/gif
          - image/jpeg
        parameters:
          - in: path
            name: attachment_file_id
            required: true
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
            description: Unique identifier of the attachment file
          - in: path
            name: file_name
            required: true
            type: string
            example: "document.pdf"
            description: Name of the file to download
        responses:
          200:
            description: Attachment file successfully downloaded
            content:
              application/octet-stream:
                schema:
                  type: string
                  format: binary
                  description: File content
        """
        attachment_file = comments_service.get_attachment_file(
            attachment_file_id
        )
        if attachment_file["comment_id"] is not None:
            comment = tasks_service.get_comment(attachment_file["comment_id"])
            user_service.check_task_access(comment["object_id"])
        elif attachment_file["chat_message_id"] is not None:
            message = chats_service.get_chat_message(
                attachment_file["chat_message_id"]
            )
            chat = chats_service.get_chat_by_id(message["chat_id"])
            entity = entities_service.get_entity(chat["object_id"])
            user_service.check_project_access(entity["project_id"])
        else:
            raise permissions.PermissionDenied()
        try:
            file_path = comments_service.get_attachment_file_path(
                attachment_file
            )
            return flask_send_file(
                file_path,
                conditional=True,
                mimetype=attachment_file["mimetype"],
                as_attachment=False,
                download_name=attachment_file["name"],
                max_age=config.CLIENT_CACHE_MAX_AGE,
                last_modified=date_helpers.get_datetime_from_string(
                    attachment_file["updated_at"]
                ),
            )
        except Exception:
            if config.LOG_FILE_NOT_FOUND:
                current_app.logger.error(
                    f"Attachment file was not found for: {attachment_file_id}"
                )
            abort(404)


class AckCommentResource(Resource):

    @jwt_required()
    def post(self, task_id, comment_id):
        """
        Acknowledge comment
        ---
        description: Acknowledge a specific comment. If it's already
          acknowledged, remove the acknowledgement.
        tags:
          - Comments
        parameters:
          - in: path
            name: task_id
            required: true
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
            description: Unique identifier of the task
          - in: path
            name: comment_id
            required: true
            type: string
            format: uuid
            example: b35b7fb5-df86-5776-b181-68564193d36
            description: Unique identifier of the comment
        responses:
          200:
            description: Comment acknowledgement status successfully updated
            content:
              application/json:
                schema:
                  type: object
                  properties:
                    id:
                      type: string
                      format: uuid
                      description: Comment unique identifier
                      example: b35b7fb5-df86-5776-b181-68564193d36
                    acknowledged:
                      type: boolean
                      description: Whether the comment is acknowledged
                      example: true
        """
        user_service.check_task_access(task_id)
        return comments_service.acknowledge_comment(comment_id)


class CommentTaskResource(Resource):

    @jwt_required()
    def post(self, task_id):
        """
        Create task comment
        ---
        description: Create a new comment for a specific task. It requires a
          text, a task_status and a person as arguments. This way, comments
          keep history of status changes. When the comment is created, it
          updates the task status with the given task status.
        tags:
          - Comments
        parameters:
          - in: path
            name: task_id
            required: true
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
            description: Unique identifier of the task
        requestBody:
          required: true
          content:
            application/json:
              schema:
                type: object
                required:
                  - task_status_id
                properties:
                  task_status_id:
                    type: string
                    format: uuid
                    description: Task status identifier
                    example: c46c8gc6-eg97-6887-c292-79675204e47
                  comment:
                    type: string
                    description: Comment text content
                    example: "This looks great! Ready for review."
                  person_id:
                    type: string
                    format: uuid
                    description: Person identifier (optional, defaults to current user)
                    example: d57d9hd7-fh08-7998-d403-80786315f58
                  created_at:
                    type: string
                    format: date-time
                    description: Creation timestamp (optional, defaults to current time)
                    example: "2023-01-01T12:00:00Z"
                  checklist:
                    type: object
                    description: Checklist items for the comment
                    example: {"item1": "Check lighting", "item2": "Verify textures"}
                  links:
                    type: array
                    items:
                      type: string
                    description: List of related links
                    example: ["https://example.com/reference1", "https://example.com/reference2"]
        responses:
          201:
            description: Comment successfully created
            content:
              application/json:
                schema:
                  type: object
                  properties:
                    id:
                      type: string
                      format: uuid
                      description: Comment unique identifier
                      example: b35b7fb5-df86-5776-b181-68564193d36
                    task_id:
                      type: string
                      format: uuid
                      description: Task identifier
                      example: a24a6ea4-ce75-4665-a070-57453082c25
                    person_id:
                      type: string
                      format: uuid
                      description: Person identifier
                      example: d57d9hd7-fh08-7998-d403-80786315f58
                    comment:
                      type: string
                      description: Comment text content
                      example: "This looks great! Ready for review."
                    task_status_id:
                      type: string
                      format: uuid
                      description: Task status identifier
                      example: c46c8gc6-eg97-6887-c292-79675204e47
                    created_at:
                      type: string
                      format: date-time
                      description: Creation timestamp
                      example: "2023-01-01T12:00:00Z"
        """
        (
            task_status_id,
            comment,
            person_id,
            created_at,
            checklist,
            links,
        ) = self.get_arguments()

        user_service.check_task_action_access(task_id)
        user_service.check_task_status_access(task_status_id)
        files = request.files

        if not permissions.has_manager_permissions():
            person_id = None
            created_at = None
        comment = comments_service.create_comment(
            person_id,
            task_id,
            task_status_id,
            comment,
            checklist,
            files,
            created_at,
            links,
        )
        return comment, 201

    def get_arguments(self):
        parser = reqparse.RequestParser()
        if request.is_json:
            location = ["values", "json"]
            parser.add_argument(
                "checklist",
                type=dict,
                action="append",
                default=[],
                location=location,
            )
            parser.add_argument(
                "links",
                type=str,
                action="append",
                default=[],
                location=location,
            )
        else:
            location = "values"
            parser.add_argument("checklist", default="[]", location=location)
            parser.add_argument("links", default="[]", location=location)
        parser.add_argument(
            "task_status_id",
            required=True,
            help="Task Status ID is missing",
            location=location,
        )
        parser.add_argument("comment", default="", location=location)
        parser.add_argument("person_id", default="", location=location)
        parser.add_argument("created_at", default="", location=location)
        args = parser.parse_args()
        return (
            args["task_status_id"],
            args["comment"],
            args["person_id"],
            args["created_at"],
            (
                args["checklist"]
                if request.is_json
                else json.loads(args["checklist"])
            ),
            (args["links"] if request.is_json else json.loads(args["links"])),
        )


class AttachmentResource(Resource):
    @jwt_required()
    def delete(self, task_id, comment_id, attachment_id):
        """
        Delete comment attachment
        ---
        description: Delete a specific attachment file linked to a comment. Only
          the comment author or project managers can delete attachments.
        tags:
          - Comments
        parameters:
          - in: path
            name: task_id
            required: true
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
            description: Unique identifier of the task
          - in: path
            name: comment_id
            required: true
            type: string
            format: uuid
            example: b35b7fb5-df86-5776-b181-68564193d36
            description: Unique identifier of the comment
          - in: path
            name: attachment_id
            required: true
            type: string
            format: uuid
            example: c46c8gc6-eg97-6887-c292-79675204e47
            description: Unique identifier of the attachment
        responses:
          204:
            description: Attachment successfully deleted
        """
        user = persons_service.get_current_user()
        comment = tasks_service.get_comment(comment_id)
        if comment["object_id"] != task_id:
            raise permissions.PermissionDenied()
        if comment["person_id"] != user["id"]:
            task = tasks_service.get_task(task_id)
            user_service.check_manager_project_access(task["project_id"])

        deletion_service.remove_attachment_file_by_id(attachment_id)
        return "", 204


class AddAttachmentToCommentResource(Resource):
    @jwt_required()
    def post(self, task_id, comment_id):
        """
        Add comment attachments
        ---
        description: Add one or more files as attachments to a specific comment.
          Supports various file types including images and documents.
        tags:
          - Comments
        consumes:
          - image/png
          - image/gif
          - image/jpeg
          - multipart/form-data
        parameters:
          - in: path
            name: task_id
            required: true
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
            description: Unique identifier of the task
          - in: path
            name: comment_id
            required: true
            type: string
            format: uuid
            example: b35b7fb5-df86-5776-b181-68564193d36
            description: Unique identifier of the comment
          - in: formData
            name: reply_id
            type: string
            format: uuid
            example: c46c8gc6-eg97-6887-c292-79675204e47
            description: Reply identifier (optional)
          - in: formData
            name: files
            type: file
            required: true
            description: Files to attach to the comment
        responses:
          201:
            description: Files successfully added as attachments
            content:
              application/json:
                schema:
                  type: array
                  items:
                    type: object
                    properties:
                      id:
                        type: string
                        format: uuid
                        description: Attachment file unique identifier
                        example: d57d9hd7-fh08-7998-d403-80786315f58
                      name:
                        type: string
                        description: File name
                        example: "document.pdf"
                      mimetype:
                        type: string
                        description: File MIME type
                        example: "application/pdf"
                      size:
                        type: integer
                        description: File size in bytes
                        example: 1024000
                      comment_id:
                        type: string
                        format: uuid
                        description: Comment identifier
                        example: b35b7fb5-df86-5776-b181-68564193d36
        """
        user = persons_service.get_current_user()
        comment = tasks_service.get_comment(comment_id)
        if comment["object_id"] != task_id:
            raise permissions.PermissionDenied()
        if comment["person_id"] != user["id"]:
            task = tasks_service.get_task(task_id)
            user_service.check_manager_project_access(task["project_id"])

        files = request.files
        comment, _ = comments_service.add_attachments_to_comment(
            comment, files, reply_id=None
        )
        return comment["attachment_files"], 201


class CommentManyTasksResource(Resource):

    @jwt_required()
    def post(self, project_id):
        """
        Create multiple comments
        ---
        description: Create several comments at once for a specific project.
          Each comment requires a text, a task id, a task_status and a person as
          arguments. This way, comments keep history of status changes. When
          the comment is created, it updates the task status with the given
          task status.
        tags:
          - Comments
        parameters:
          - in: path
            name: project_id
            required: true
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
            description: Unique identifier of the project
        requestBody:
          required: true
          content:
            application/json:
              schema:
                type: array
                items:
                  type: object
                  required:
                    - task_status_id
                    - object_id
                  properties:
                    task_status_id:
                      type: string
                      format: uuid
                      description: Task status identifier
                      example: c46c8gc6-eg97-6887-c292-79675204e47
                    comment:
                      type: string
                      description: Comment text content
                      example: "This looks great! Ready for review."
                    person_id:
                      type: string
                      format: uuid
                      description: Person identifier (optional, defaults to current user)
                      example: d57d9hd7-fh08-7998-d403-80786315f58
                    object_id:
                      type: string
                      format: uuid
                      description: Task identifier
                      example: e68e0ie8-gi19-8009-e514-91897426g69
                    created_at:
                      type: string
                      format: date-time
                      description: Creation timestamp (optional, defaults to current time)
                      example: "2023-01-01T12:00:00Z"
                    checklist:
                      type: object
                      description: Checklist items for the comment
                      example: {"item1": "Check lighting", "item2": "Verify textures"}
                    links:
                      type: array
                      items:
                        type: string
                      description: List of related links
                      example: ["https://example.com/reference1", "https://example.com/reference2"]
        responses:
          201:
            description: Comments successfully created
            content:
              application/json:
                schema:
                  type: array
                  items:
                    type: object
                    properties:
                      id:
                        type: string
                        format: uuid
                        description: Comment unique identifier
                        example: b35b7fb5-df86-5776-b181-68564193d36
                      task_id:
                        type: string
                        format: uuid
                        description: Task identifier
                        example: e68e0ie8-gi19-8009-e514-91897426g69
                      person_id:
                        type: string
                        format: uuid
                        description: Person identifier
                        example: d57d9hd7-fh08-7998-d403-80786315f58
                      comment:
                        type: string
                        description: Comment text content
                        example: "This looks great! Ready for review."
                      task_status_id:
                        type: string
                        format: uuid
                        description: Task status identifier
                        example: c46c8gc6-eg97-6887-c292-79675204e47
                      created_at:
                        type: string
                        format: date-time
                        description: Creation timestamp
                        example: "2023-01-01T12:00:00Z"
        """
        comments = request.json
        if not isinstance(comments, list):
            abort(400, "Request body must be a JSON array.")
        person = persons_service.get_current_user(relations=True)
        try:
            user_service.check_manager_project_access(project_id)
        except permissions.PermissionDenied:
            comments = self.get_allowed_comments_only(comments, person)
        result = []
        for comment in comments:
            if (
                "task_status_id" not in comment
                or "object_id" not in comment
                or "comment" not in comment
            ):
                continue
            user_service.check_task_status_access(
                comment["task_status_id"]
            )
            comment = comments_service.create_comment(
                person["id"],
                comment["object_id"],
                comment["task_status_id"],
                comment["comment"],
                [],
                {},
                None,
                comment.get("links", []),
            )
            result.append(comment)
        return result, 201

    def get_allowed_comments_only(self, comments, person):
        allowed_comments = []
        for comment in comments:
            try:
                task = tasks_service.get_task(
                    comment["object_id"], relations=True
                )
                if (
                    person["role"] == "supervisor"
                    and (
                        len(person["departments"]) == 0
                        or tasks_service.get_task_type(task["task_type_id"])[
                            "department_id"
                        ]
                        in person["departments"]
                    )
                ) or person["id"] in task["assignees"]:
                    allowed_comments.append(comment)
            except permissions.PermissionDenied:
                pass
            except KeyError:
                pass
        return allowed_comments


class ReplyCommentResource(Resource, ArgsMixin):

    @jwt_required()
    def post(self, task_id, comment_id):
        """
        Reply to comment
        ---
        description: Add a reply to a specific comment. The reply will be added
          to the comment's replies list.
        tags:
          - Comments
        parameters:
          - in: path
            name: task_id
            required: true
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
            description: Unique identifier of the task
          - in: path
            name: comment_id
            required: true
            type: string
            format: uuid
            example: b35b7fb5-df86-5776-b181-68564193d36
            description: Unique identifier of the comment
          - in: formData
            name: text
            type: string
            example: "Thanks for the feedback!"
            description: Reply text content
        responses:
          200:
            description: Reply successfully added to comment
            content:
              application/json:
                schema:
                  type: object
                  properties:
                    id:
                      type: string
                      format: uuid
                      description: Reply unique identifier
                      example: c46c8gc6-eg97-6887-c292-79675204e47
                    comment_id:
                      type: string
                      format: uuid
                      description: Parent comment identifier
                      example: b35b7fb5-df86-5776-b181-68564193d36
                    text:
                      type: string
                      description: Reply text content
                      example: "Thanks for the feedback!"
                    person_id:
                      type: string
                      format: uuid
                      description: Person identifier who made the reply
                      example: d57d9hd7-fh08-7998-d403-80786315f58
                    created_at:
                      type: string
                      format: date-time
                      description: Creation timestamp
                      example: "2023-01-01T12:00:00Z"
        """
        comment = tasks_service.get_comment(comment_id)
        if comment["object_id"] != task_id:
            raise permissions.PermissionDenied()
        current_user = persons_service.get_current_user()
        if comment["person_id"] != current_user["id"]:
            if permissions.has_client_permissions():
                author = persons_service.get_person(comment["person_id"])
                if (
                    current_user["studio_id"] != author["studio_id"]
                    and author["role"] == "client"
                ):
                    raise permissions.PermissionDenied()
            user_service.check_task_action_access(task_id)

        args = self.get_args(
            [
                ("text", "", False),
            ]
        )
        files = request.files
        return comments_service.reply_comment(
            comment_id, args["text"], files=files
        )


class DeleteReplyCommentResource(Resource):

    @jwt_required()
    def delete(self, task_id, comment_id, reply_id):
        """
        Delete comment reply
        ---
        description: Delete a specific reply from a comment.
         Only the reply author or administrators can delete replies.
        tags:
          - Comments
        parameters:
          - in: path
            name: task_id
            required: true
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
            description: Unique identifier of the task
          - in: path
            name: comment_id
            required: true
            type: string
            format: uuid
            example: b35b7fb5-df86-5776-b181-68564193d36
            description: Unique identifier of the comment
          - in: path
            name: reply_id
            required: true
            type: string
            format: uuid
            example: c46c8gc6-eg97-6887-c292-79675204e47
            description: Unique identifier of the reply
        responses:
          200:
            description: Reply successfully deleted
        """
        reply = comments_service.get_reply(comment_id, reply_id)
        current_user = persons_service.get_current_user()
        if reply["person_id"] != current_user["id"]:
            permissions.check_admin_permissions()
        return comments_service.delete_reply(comment_id, reply_id)


class ProjectAttachmentFiles(Resource):

    @jwt_required()
    def get(self, project_id):
        """
        Get project attachment files
        ---
        description: Retrieve all attachment files related to a specific
          project. Requires administrator permissions.
        tags:
          - Comments
        parameters:
          - in: path
            name: project_id
            required: true
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
            description: Unique identifier of the project
        responses:
          200:
            description: Project attachment files successfully retrieved
            content:
              application/json:
                schema:
                  type: array
                  items:
                    type: object
                    properties:
                      id:
                        type: string
                        format: uuid
                        description: Attachment file unique identifier
                        example: d57d9hd7-fh08-7998-d403-80786315f58
                      name:
                        type: string
                        description: File name
                        example: "document.pdf"
                      mimetype:
                        type: string
                        description: File MIME type
                        example: "application/pdf"
                      size:
                        type: integer
                        description: File size in bytes
                        example: 1024000
                      comment_id:
                        type: string
                        format: uuid
                        description: Comment identifier
                        example: b35b7fb5-df86-5776-b181-68564193d36
                      project_id:
                        type: string
                        format: uuid
                        description: Project identifier
                        example: a24a6ea4-ce75-4665-a070-57453082c25
        """
        permissions.check_admin_permissions()
        return comments_service.get_all_attachment_files_for_project(
            project_id
        )


class TaskAttachmentFiles(Resource):

    @jwt_required()
    def get(self, task_id):
        """
        Get task attachment files
        ---
        description: Retrieve all attachment files related to a specific task.
          Requires administrator permissions.
        tags:
          - Comments
        parameters:
          - in: path
            name: task_id
            required: true
            type: string
            format: uuid
            example: a24a6ea4-ce75-4665-a070-57453082c25
            description: Unique identifier of the task
        responses:
          200:
            description: Task attachment files successfully retrieved
            content:
              application/json:
                schema:
                  type: array
                  items:
                    type: object
                    properties:
                      id:
                        type: string
                        format: uuid
                        description: Attachment file unique identifier
                        example: d57d9hd7-fh08-7998-d403-80786315f58
                      name:
                        type: string
                        description: File name
                        example: "document.pdf"
                      mimetype:
                        type: string
                        description: File MIME type
                        example: "application/pdf"
                      size:
                        type: integer
                        description: File size in bytes
                        example: 1024000
                      comment_id:
                        type: string
                        format: uuid
                        description: Comment identifier
                        example: b35b7fb5-df86-5776-b181-68564193d36
                      task_id:
                        type: string
                        format: uuid
                        description: Task identifier
                        example: a24a6ea4-ce75-4665-a070-57453082c25
                      reply_id:
                        type: string
                        format: uuid
                        description: Reply identifier if attached to a reply
                        example: c46c8gc6-eg97-6887-c292-79675204e47
        """
        permissions.check_admin_permissions()
        return comments_service.get_all_attachment_files_for_task(task_id)
